# -*- coding: UTF-8 -*-
"""
Actions
=======
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Extra actions used for manipulating the browser's behavior.
"""

from typing import Union

try:
    from typing import Sequence
    from typing import Tuple
except ImportError:
    from collections.abc import Sequence
    from builtins import tuple as Tuple

from typing_extensions import Literal

from dash.testing.composite import DashComposite
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.common.by import By

from .utils import wait_for, attribute_value_neq


__all__ = (
    "drag_and_drop",
    "click_relatively",
    "get_input_element",
    "get_input_box",
    "get_input_del_button",
    "get_select_element",
    "get_select_expand",
    "get_select_del_button",
    "get_select_control_points",
    "get_select_item_by_label",
    "get_select_current_item",
    "get_canvas_image",
)


def drag_and_drop(
    dash_duo: DashComposite,
    element: WebElement,
    pos_start: Tuple[int, int] = (0, 0),
    pos_end: Tuple[int, int] = (0, 0),
) -> None:
    """Emulator behavior: Drag and drop.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    element: `WebElement`
        The anchor of this event. The positions will be calculated as relative
        positions of the upper left corner of this element.

    pos_start: `(int, int)`
        The starting position. The mouse will press since here.

    pos_end: `(int, int)`
        The ending position. The mouse will release at here.
    """
    driver: WebDriver = dash_duo.driver
    action = ActionChains(driver)
    action.move_to_element_with_offset(element, pos_start[0], pos_start[1])
    action.click_and_hold()
    action.move_to_element_with_offset(element, pos_end[0], pos_end[1])
    action.release()
    action.perform()


def click_relatively(
    dash_duo: DashComposite, element: WebElement, pos: Tuple[int, int] = (0, 0)
) -> None:
    """Emulator behavior: Click a relative position of an element.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    element: `WebElement`
        The anchor of this event. The positions will be calculated as relative
        positions of the upper left corner of this element.

    pos: `(int, int)`
        The position where the mouse will click.
    """
    driver: WebDriver = dash_duo.driver
    action = ActionChains(driver)
    action.move_to_element_with_offset(element, pos[0], pos[1])
    action.click()
    action.perform()


def get_input_element(
    dash_duo: DashComposite, annotator: Union[str, WebElement]
) -> WebElement:
    """Emulator behavior: Get the `<div class="rp-selected-input">` element of a
    selected annotation.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `WebElement`
        The `<div class="rp-selected-input">` element in the currently selected
        annotation.
    """
    if isinstance(annotator, str):
        dash_duo.wait_for_element_by_css_selector(
            "{0} > .rp-stage > .rp-selected-input".format(annotator),
        )
        element: WebElement = dash_duo.find_element(
            "{0} > .rp-stage > .rp-selected-input".format(annotator),
        )
    else:
        wait_for(
            dash_duo,
            lambda: annotator.find_element(
                By.CSS_SELECTOR, ".rp-stage > .rp-selected-input"
            ),
        )
        element: WebElement = annotator.find_element(
            By.CSS_SELECTOR,
            ".rp-stage > .rp-selected-input",
        )
    return element


def get_input_box(
    dash_duo: DashComposite, annotator: Union[str, WebElement]
) -> WebElement:
    """Emulator behavior: Get the `<input>` element of a selected annotation.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `WebElement`
        The `<input>` element in the currently selected annotation.
    """
    return get_input_element(dash_duo, annotator).find_element(By.TAG_NAME, "input")


def get_input_del_button(
    dash_duo: DashComposite, annotator: Union[str, WebElement]
) -> WebElement:
    """Emulator behavior: Get the delete button, i.e. the
    `<div class="rp-default-input-section_delete">` element of a selected annotation.

    This element appears only when the `<input>` element is available. It should appear
    near the `<input>` element.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `WebElement`
        The `<div class="rp-default-input-section_delete">` element in the currently
        selected annotation.
    """
    return get_input_element(dash_duo, annotator).find_element(
        By.CSS_SELECTOR, ".rp-default-input-section_delete"
    )


def get_select_element(
    dash_duo: DashComposite, annotator: Union[str, WebElement]
) -> WebElement:
    """Emulator behavior: Get the React-selector, i.e. the
    `<div class="dpa-dropdown...">` element of a selected annotation.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `WebElement`
        The selector element in the currently selected annotation.
    """
    if isinstance(annotator, str):
        dash_duo.wait_for_element_by_css_selector(
            "{0} > .rp-stage > .rp-selected-input > "
            "div[class^='dpa-dropdown']".format(annotator),
        )
        element: WebElement = dash_duo.find_element(
            "{0} > .rp-stage > .rp-selected-input > "
            "div[class^='dpa-dropdown']".format(annotator)
        )
    else:
        wait_for(
            dash_duo,
            lambda: annotator.find_element(
                By.CSS_SELECTOR,
                ".rp-stage > .rp-selected-input > div[class^='dpa-dropdown']",
            ),
        )
        element: WebElement = annotator.find_element(
            By.CSS_SELECTOR,
            ".rp-stage > .rp-selected-input > div[class^='dpa-dropdown']",
        )
    return element


def get_select_expand(
    dash_duo: DashComposite, annotator: Union[str, WebElement]
) -> WebElement:
    """Emulator behavior: Get the control region of the selector. The control region
    is an element that will trigger the dropdown menu when being clicked.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `WebElement`
        The control region of the selector element in the currently selected
        annotation.
    """
    return get_select_element(dash_duo, annotator).find_element(
        By.CSS_SELECTOR, "div[class$='-control']"
    )


def get_select_del_button(
    dash_duo: DashComposite, annotator: Union[str, WebElement]
) -> WebElement:
    """Emulator behavior: Get the delete button of the selector.

    This element appears only when the selector element is available. It should appear
    near the control region of the selector element.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `WebElement`
        The delete button of the selector element in the currently selected
        annotation.
    """
    return get_select_element(dash_duo, annotator).find_element(
        By.CSS_SELECTOR, "div[class^='dpa-btn-delete']"
    )


def get_select_control_points(
    dash_duo: DashComposite, annotator: Union[str, WebElement]
) -> Sequence[WebElement]:
    """Emulator behavior: Get the control points of the selector. The control points
    represent icons in the selector that will trigger special behaviors. For example,
    a "clear" icon will make the currently selected item cleared.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `[WebElement]`
        All clickable control point icons detected in the selector.
    """
    return get_select_element(dash_duo, annotator).find_elements(
        By.CSS_SELECTOR,
        "div[class$='-control'] > div:nth-child(2) > div[class$='-indicatorContainer']",
    )


def get_select_item_by_label(
    dash_duo: DashComposite, annotator: Union[str, WebElement], item_label: str
) -> WebElement:
    """Emulator behavior: Get a selectable item from the menu by checking its label.

    This element can be located only when the dropdown menu is activated by clicking
    the control region of the selector.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `WebElement`
        The selectable item in the menu. Click this value will make the currently
        selected item switched to this one.
    """
    element: WebElement = get_select_element(dash_duo, annotator).find_element(
        By.CSS_SELECTOR, "div[class$='-menu']"
    )
    element: WebElement = element.find_element(
        By.XPATH, "//*[text()='{0}']".format(item_label)
    )
    return element


def get_select_current_item(
    dash_duo: DashComposite,
    annotator: Union[str, WebElement],
    expect_to_be: Literal["item", "placeholder"] = "item",
) -> WebElement:
    """Emulator behavior: Get the currently selected item. The selected item is an
    element in the control region of the selector. Its `text` property is the label
    of the currently selected item.

    If the element cannot be found, will throw a timeout error.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    Returns
    -------
    #1: `WebElement`
        The currently selected item of the selector element in the currently selected
        annotation.
    """
    element = get_select_element(dash_duo, annotator)
    if expect_to_be == "item":
        wait_for(
            dash_duo,
            lambda: (
                isinstance(
                    element.find_element(
                        By.CSS_SELECTOR,
                        "div[class$='-control'] > div > div[class$=-singleValue]",
                    ),
                    WebElement,
                )
            ),
        )
        element: WebElement = element.find_element(
            By.CSS_SELECTOR, "div[class$='-control'] > div > div[class$=-singleValue]"
        )
    elif expect_to_be == "placeholder":
        wait_for(
            dash_duo,
            lambda: isinstance(
                element.find_element(
                    By.CSS_SELECTOR,
                    "div[class$='-control'] > div > div[id$=-placeholder]",
                ),
                WebElement,
            ),
        )
        element: WebElement = element.find_element(
            By.CSS_SELECTOR, "div[class$='-control'] > div > div[id$=-placeholder]"
        )
    else:
        raise TypeError(
            'The arugment "expect_to_be" is not recognizable: {0}'.format(expect_to_be)
        )
    return element


def get_canvas_image(
    dash_duo: DashComposite,
    annotator: Union[str, WebElement],
    canvas_selector: Literal[".rp-image", ".rp-shapes"] = ".rp-image",
) -> str:
    """Emulator behavior: Get the canvas image.

    This method will returns a base64-encoded and png-format snapshot of a canvas.
    This snapshot can be processed by the image utilities to validate the visualized
    results.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator.

    annotator: `str | WebElement`
        The CSS-selector or the element of the annotator. To obtain the annotator
        element, the following code can be used:
        ``` python
        dash_duo.find_element("#annotator")
        ```

    canvas_selector: `".rp-image" | ".rp-shapes"`
        The CSS-selector used for locating the canvas. `".rp-image"` and `".rp-shapes"`
        represent the displayed image and the annotations, respectively.

    Returns
    -------
    #1: `str`
        The base64-encoded and png-format snapshot of the selected canvas.
    """
    driver: WebDriver = dash_duo.driver
    element: WebElement = (
        annotator
        if isinstance(annotator, WebElement)
        else dash_duo.find_element(annotator)
    )
    dash_duo._wait_for(
        attribute_value_neq(element, "data-dash-is-loading", "true"),
        timeout=None,
        msg=(
            "timeout {0}s => waiting for the element {1} to be loaded.".format(
                dash_duo._wait_timeout,
                (
                    annotator
                    if isinstance(annotator, str)
                    else (element.tag_name + element.id)
                ),
            )
        ),
    )
    selector = ".rp-stage > {0}".format(canvas_selector)
    wait_for(
        dash_duo,
        lambda: isinstance(element.find_element(By.CSS_SELECTOR, selector), WebElement),
    )
    element = element.find_element(By.CSS_SELECTOR, selector)
    assert isinstance(element, WebElement)
    res = str(
        driver.execute_script(
            "return {data: arguments[0].toDataURL('image/png')}", element
        ).get("data", "")
    )
    assert res.startswith("data:image/png;base64,")
    return res



# -*- coding: UTF-8 -*-
"""
Utilities
=========
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Extra functionalities used for enhancing the tests.
"""

import os
import collections.abc

from typing import Optional, Any

try:
    from typing import Sequence, Mapping, Callable
except ImportError:
    from collections.abc import Sequence, Mapping, Callable

from dash.testing.composite import DashComposite
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.common.exceptions import StaleElementReferenceException


__all__ = (
    "is_eq",
    "is_eq_mapping",
    "is_eq_sequence",
    "is_mapping_with_keys",
    "tst_folder",
    "asset_folder",
    "wait_func",
    "attribute_value_to_equal",
    "attribute_value_neq",
    "wait_for_the_attribute_value_to_equal",
    "wait_for_the_attribute_value_neq",
    "wait_for_dcc_loading",
    "wait_for",
)


def is_eq(val: Any, ref: Any) -> bool:
    """Safely check whether `val == ref`"""
    if isinstance(ref, (str, bytes)):
        return isinstance(val, ref.__class__) and val == ref
    if isinstance(ref, collections.abc.Sequence):
        return is_eq_sequence(val, ref)
    elif isinstance(ref, collections.abc.Mapping):
        return is_eq_mapping(val, ref)
    else:
        return isinstance(val, ref.__class__) and val == ref


def is_eq_mapping(val: Any, ref: Mapping[Any, Any]) -> bool:
    """Safely check whether `val == ref`, where `ref` is a mapping."""
    if not isinstance(val, collections.abc.Mapping):
        return False
    return val == ref


def is_eq_sequence(val: Any, ref: Sequence[Any]) -> bool:
    """Safely check whether `val == ref`, where `ref` is a sequence."""
    if isinstance(val, (str, bytes)) or (not isinstance(val, collections.abc.Sequence)):
        return False
    return tuple(val) == tuple(ref)


def is_mapping_with_keys(val: Any, keys: Sequence[Any]) -> bool:
    """Check whether `val` is mapping and this mapping has keys specified by `keys`.

    If `keys` is not a sequence, will treat it as one key.
    """
    if isinstance(keys, (str, bytes)) or (
        not isinstance(keys, collections.abc.Sequence)
    ):
        keys = (keys,)
    if not isinstance(val, collections.abc.Mapping):
        return False
    return set(val.keys()) == set(keys)


def tst_folder(file_name: Optional[str] = None) -> str:
    """Get the path of the testing folder.

    Arguments
    ---------
    file_name: `str`
        If specified, will view this value as the name of the file in the testing
        folder.

    Returns
    -------
    #1: `str`
        If `file_name` is provided, return the path of that file.
        Otherwise, return the testing folder path.
    """
    path = os.path.dirname(__file__)
    return (
        os.path.join(path, file_name)
        if isinstance(file_name, str) and file_name
        else path
    )


def asset_folder(file_name: Optional[str] = None) -> str:
    """Get the path of the asset folder.

    Arguments
    ---------
    file_name: `str`
        If specified, will view this value as the name of the file in the asset folder.

    Returns
    -------
    #1: `str`
        If `file_name` is provided, return the path of that file.
        Otherwise, return the asset folder path.
    """
    path = os.path.join(os.path.dirname(__file__), "assets")
    return (
        os.path.join(path, file_name)
        if isinstance(file_name, str) and file_name
        else path
    )


class wait_func:
    """Wait-for method: text value does not equal to something.

    The instance of this class serves as a method used by `DashComposite._waitfor`.
    It will listen to the state of the chosen element until its specific attribute
    value is not the specified value any more.
    """

    def __init__(self, func: Callable[[], bool]):
        """Initialization.

        Arguments
        ---------
        func: `() -> bool`
            The function to be waited. If `func()` returns `True`, the waiting will be
            finished.
        """
        self.func = func

    def __call__(self, driver):
        try:
            return self.func()
        except StaleElementReferenceException:
            return False


class attribute_value_to_equal:
    """Wait-for method: attribute value equals to something.

    The instance of this class serves as a method used by `DashComposite._waitfor`.
    It will listen to the state of the chosen element until its specific attribute
    value is not the specified value any more.
    """

    def __init__(self, element: WebElement, attribute: str, value: Any) -> None:
        """Initialization.

        Arguments
        ---------
        element: `WebElement`
            The selected selenium `WebElement` where the attribtue will be listened to.

        attribtue: `str`
            The attribute name to be checked. Normally, this name should starts with
            `"data-"`.

        value: `Any`
            The value that the attribute expects to be. Normally, this value
            should be a string.
        """
        self.element = element
        self.attribute = attribute
        self.value = value
        self.value_type = type(value)

    def __call__(self, driver: WebDriver):
        """Wait-for method."""
        try:
            element_attribute = self.element.get_attribute(self.attribute)
            return isinstance(element_attribute, self.value_type) and (
                element_attribute == self.value
            )
        except StaleElementReferenceException:
            return False


class attribute_value_neq:
    """Wait-for method: attribute value does not equal to something.

    The instance of this class serves as a method used by `DashComposite._waitfor`.
    It will listen to the state of the chosen element until its specific attribute
    value is not the specified value any more.
    """

    def __init__(self, element: WebElement, attribute: str, value: Any) -> None:
        """Initialization.

        Arguments
        ---------
        element: `WebElement`
            The selected selenium `WebElement` where the attribtue will be listened to.

        attribtue: `str`
            The attribute name to be checked. Normally, this name should starts with
            `"data-"`.

        value: `Any`
            The value that the attribute needs to quit from. Normally, this value
            should be a string.
        """
        self.element = element
        self.attribute = attribute
        self.value = value
        self.value_type = type(value)

    def __call__(self, driver: WebDriver):
        """Wait-for method."""
        try:
            element_attribute = self.element.get_attribute(self.attribute)
            return (not isinstance(element_attribute, self.value_type)) or (
                element_attribute != self.value
            )
        except StaleElementReferenceException:
            return False


def wait_for(
    dash_duo: DashComposite,
    func: Optional[Callable[[], bool]] = None,
    timeout: Optional[int] = None,
) -> None:
    """The general `wait_for` method for `dash_duo`.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator providing the `_wait_for` method.

    func: `(() -> bool) | None`
        The validator function. The input value is the attribute value. If this
        function returns `True`, the check will pass.

        If this function is not specified, will not wait for anything and return
        immediately.

    timeout: `int | None`
        The customized time out (seconds) length that this method needs to wait.
    """
    if func is None:
        return
    dash_duo._wait_for(
        wait_func(func),
        timeout=timeout,
        msg=(
            "timeout {0}s => waiting for the function {1}.".format(
                timeout or dash_duo._wait_timeout, str(func)
            )
        ),
    )


def wait_for_the_attribute_value_to_equal(
    dash_duo: DashComposite,
    selector: str,
    by: str = "CSS_SELECTOR",
    attribute: str = "data-any",
    value: Any = "",
    timeout: Optional[int] = None,
) -> None:
    """Select an element, and wait until its attribute equals to the specific value.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator providing the `_wait_for` method.

    selector: `str`
        The selector used for locating the target element.

    by: `str`
        The method of using the selector.
        Valid values: "CSS_SELECTOR", "ID", "NAME", "TAG_NAME",
        "CLASS_NAME", "LINK_TEXT", "PARTIAL_LINK_TEXT", "XPATH".

    attribtue: `str`
        The attribute name to be checked. Normally, this name should starts with
        `"data-"`.

    value: `Any`
        The value that the attribute expects to be. Normally, this value should be a
        string.

    timeout: `int | None`
        The customized time out (seconds) length that this method needs to wait.
    """
    dash_duo._wait_for(
        attribute_value_to_equal(dash_duo.find_element(selector, by), attribute, value),
        timeout=timeout,
        msg=(
            "timeout {0}s => waiting for the element {1} until the attribute {2} is "
            "not {3}.".format(
                timeout or dash_duo._wait_timeout, selector, attribute, value
            )
        ),
    )


def wait_for_the_attribute_value_neq(
    dash_duo: DashComposite,
    selector: str,
    by: str = "CSS_SELECTOR",
    attribute: str = "data-any",
    value: Any = "",
    timeout: Optional[int] = None,
) -> None:
    """Select an element, and wait until its attribute does not equal to the specific
    value.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator providing the `_wait_for` method.

    selector: `str`
        The selector used for locating the target element.

    by: `str`
        The method of using the selector.
        Valid values: "CSS_SELECTOR", "ID", "NAME", "TAG_NAME",
        "CLASS_NAME", "LINK_TEXT", "PARTIAL_LINK_TEXT", "XPATH".

    attribtue: `str`
        The attribute name to be checked. Normally, this name should starts with
        `"data-"`.

    value: `Any`
        The value that the attribute needs to quit from. Normally, this value should
        be a string.

    timeout: `int | None`
        The customized time out (seconds) length that this method needs to wait.
    """
    dash_duo._wait_for(
        attribute_value_neq(dash_duo.find_element(selector, by), attribute, value),
        timeout=None,
        msg=(
            "timeout {0}s => waiting for the element {1} until the attribute {2} is "
            "not {3}.".format(
                timeout or dash_duo._wait_timeout, selector, attribute, value
            )
        ),
    )


def wait_for_dcc_loading(
    dash_duo: DashComposite,
    selector: str,
    by: str = "CSS_SELECTOR",
    timeout: Optional[int] = None,
) -> None:
    """Select an element, and wait until it quits from the is-loading state.

    Arguments
    ---------
    dash_duo: `DashComposite`
        The dash emulator providing the `_wait_for` method.

    selector: `str`
        The selector used for locating the target element.

    by: `str`
        The method of using the selector.
        Valid values: "CSS_SELECTOR", "ID", "NAME", "TAG_NAME",
        "CLASS_NAME", "LINK_TEXT", "PARTIAL_LINK_TEXT", "XPATH".

    timeout: `int | None`
        The customized time out (seconds) length that this method needs to wait.
    """
    dash_duo._wait_for(
        attribute_value_neq(
            dash_duo.find_element(selector, by), "data-dash-is-loading", "true"
        ),
        timeout=None,
        msg=(
            "timeout {0}s => waiting for the element {1} to be loaded.".format(
                timeout or dash_duo._wait_timeout, selector
            )
        ),
    )



# -*- coding: UTF-8 -*-
"""
Tests
=====
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Testing codes powered by `pytest`. This namespace module is used for providing the
unit tests for the package.
"""



# -*- coding: UTF-8 -*-
"""
Sanitize
========
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
The tests for sanitizing and querying the annotation data.
"""

import json
import logging

try:
    from typing import Generator
except ImportError:
    from collections.abc import Generator

import pytest

import dash_picture_annotation as dpa

from .utils import tst_folder


__all__ = ("TestSanitize",)


class TestSanitize:
    """Test the sanitizing the querying of the annotation data."""

    @pytest.fixture(scope="class")
    def data(self) -> Generator[dpa.Annotations, None, None]:
        """Fixture of the testing data."""
        log = logging.getLogger("dash_picture_annotation.test")
        file_name = "data-sanitize"
        with open(tst_folder("{0}.json".format(file_name)), "r") as fobj:
            _data = json.load(fobj)
        log.info("Use testing data: {0}".format(file_name))
        yield _data

    def test_sanitize_get_data(self, data: dpa.Annotations) -> None:
        """Test the methods for getting data items."""
        log = logging.getLogger("dash_picture_annotation.test")

        anno = dpa.get_data_item(data, "ZxdA2p")
        assert anno["id"] == "ZxdA2p"
        assert abs(anno["mark"]["x"] - 152.14361702127658) < 1e-3
        log.info("Get data: {0}".format(anno))

        with pytest.raises(KeyError):
            dpa.get_data_item(data, "not-exist")

        anno = dpa.get_data_item_with_default(data, "not-exist")
        assert anno is None

        annos = dpa.get_data_items(data, "type-1")
        assert len(annos) == 0
        log.info("Get data: {0}".format(annos))

        annos = dpa.get_data_items(data, "type-2")
        assert len(annos) == 1
        assert annos[0].get("comment") == "type-2"
        log.info("Get data: {0}".format(annos))

        annos = dpa.get_data_items(data, "type-7")
        assert len(annos) == 2
        anno = dpa.get_data_item(annos, "yJwGxK")
        assert anno.get("comment") == "type-7"
        log.info("Get data: {0}".format(annos))

        annos = dpa.get_data_items_by_regex(data, r"type-[678]")
        assert len(annos) == 3
        anno = dpa.get_data_item(annos, "yJwGxK")
        assert anno.get("comment") in ("type-6", "type-7", "type-8")
        log.info("Get data: {0}".format(annos))

    def test_sanitize_all_info(self, data: dpa.Annotations) -> None:
        """Test the methods for gathering high-level information."""
        log = logging.getLogger("dash_picture_annotation.test")

        anno_ids = dpa.get_all_ids(data)
        assert not set(anno_ids).difference(
            ("ZxdA2p", "hCC6Gi", "yJwGxK", "EWCEpN", "3Pb3rh")
        )
        anno_ids_2 = dpa.get_all_ids(data["data"])
        assert not set(anno_ids).difference(set(anno_ids_2))
        log.info("Get all IDs: {0}".format(anno_ids))

        anno_comments = dpa.get_all_comments(data)
        assert not anno_comments.difference(("type-2", "type-6", "type-7", ""))
        anno_comments_2 = dpa.get_all_comments(data["data"])
        assert not set(anno_comments).difference(anno_comments_2)
        log.info("Get all comments: {0}".format(anno_comments))

    def test_sanitize_compare(self, data: dpa.Annotations) -> None:
        """Test the methods for gathering high-level information."""
        log = logging.getLogger("dash_picture_annotation.test")

        anno_0 = data["data"][0]
        anno_1 = data["data"][1]

        assert dpa.compare_anno_marks(anno_0, anno_0)
        assert dpa.compare_anno_marks(anno_1, anno_1)
        assert not dpa.compare_anno_marks(anno_0, anno_1)
        log.info("Distinguish the difference two annotations.")

        assert dpa.compare_anno_marks(anno_0["mark"], anno_0)
        assert dpa.compare_anno_marks(anno_1, anno_1["mark"])
        assert not dpa.compare_anno_marks(anno_0["mark"], anno_1["mark"])
        log.info("Confirm the compatibility of using marks directly.")

        anno_mark = anno_0["mark"].copy()
        anno_mark["width"] = -anno_mark["width"]
        assert not dpa.compare_anno_marks(anno_0, anno_mark)
        log.info("Negative width causes the mark to move.")

        anno_mark["x"] = anno_mark["x"] - anno_mark["width"]
        assert dpa.compare_anno_marks(anno_0, anno_mark)
        log.info("Correct the mark poistion back by changing x.")

        anno_copy = anno_0.copy()
        anno_copy["mark"] = anno_copy["mark"].copy()
        anno_copy["comment"] = "changed-comment"
        assert dpa.compare_anno_marks(anno_0, anno_copy)
        log.info("Change comment does not influence the comparison.")

    def test_sanitize_data(self, data: dpa.Annotations) -> None:
        """Test the methods for gathering high-level information."""
        log = logging.getLogger("dash_picture_annotation.test")

        anno_bad = data["data"][4]
        assert anno_bad["id"] == "yJwGxK"
        assert anno_bad["mark"]["width"] < 0
        log.info("Get the bad item: {0}".format(anno_bad))

        anno_good = dpa.sanitize_data_item(anno_bad)
        assert anno_good is not None
        assert dpa.compare_anno_marks(anno_bad, anno_good)
        log.info("Get the sanitized item: {0}".format(anno_good))

        assert anno_bad["mark"]["width"] < 0
        assert anno_good["mark"]["width"] > 0
        log.info("Confirm that the bad item is not changed: {0}".format(anno_bad))

        bad_ids = dpa.get_all_ids(data)
        bad_comments = dpa.get_all_comments(data)
        assert not set(bad_ids).difference(
            ("ZxdA2p", "hCC6Gi", "yJwGxK", "EWCEpN", "3Pb3rh")
        )
        assert not bad_comments.difference(("type-2", "type-6", "type-7", ""))
        log.info("IDs of bad data: {0}".format(bad_ids))
        log.info("Comments of bad data: {0}".format(bad_comments))

        s_data = dpa.sanitize_data(data, deduplicate="add")
        good_ids = dpa.get_all_ids(s_data)
        good_comments = dpa.get_all_comments(s_data)
        log.info("IDs of sanitized data: {0}".format(good_ids))
        log.info("Comments of sanitized data: {0}".format(good_comments))

        assert (
            not set(good_ids)
            .intersection(bad_ids)
            .difference(("ZxdA2p", "hCC6Gi", "yJwGxK", "EWCEpN", "3Pb3rh"))
        )
        assert len(set(good_ids)) == len(s_data["data"])
        log.info("Confirm that the IDs are sanitized.")

        assert not good_comments.difference(bad_comments)
        log.info("Confirm that the comments are not changed in the sanitized data.")

        _bad_ids = dpa.get_all_ids(data)
        _bad_comments = dpa.get_all_comments(data)
        assert not set(_bad_ids).difference(bad_ids)
        assert not bad_comments.difference(_bad_comments)
        log.info("Confirm that the bad data is not changed: {0}".format(anno_bad))

        _s_data = dpa.sanitize_data(data, deduplicate="drop")
        log.info(
            "Use the drop mode to sanitize data: len(sdata_drop_mode)={0} < {1}="
            "len(sdata_add_mode)".format(len(_s_data["data"]), len(s_data["data"]))
        )
        assert len(_s_data["data"]) < len(s_data["data"])



# -*- coding: UTF-8 -*-
"""
Minimal
=======
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
The tests for the `minimal.py` application. These tests will run a browser emulator
powered by `selenium` and `dash.testing`. The basic functionalities of the demo
will be checked one by one.
"""

import json
import logging

try:
    from typing import Generator
except ImportError:
    from collections.abc import Generator

import pytest

import dash
from dash.testing.application_runners import import_app
from dash.testing.composite import DashComposite
from selenium.webdriver.remote.webelement import WebElement

from .utils import wait_for_dcc_loading
from .actions import (
    drag_and_drop,
    click_relatively,
    get_input_box,
    get_input_del_button,
)


__all__ = ("TestMinimal",)


@pytest.mark.with_dash
class TestMinimal:
    """Test the rendering the Dash app: Minimal"""

    @pytest.fixture(scope="class")
    def dash_app(self) -> Generator[dash.Dash, None, None]:
        log = logging.getLogger("dash_picture_annotation.test")
        log.info("Initialize the Dash app.")
        app = import_app("examples.minimal")
        yield app
        log.info("Remove the Dash app.")
        del app

    def test_minimal_create_new_annotation(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the minimal app and create a new annotation box.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        output: WebElement = dash_duo.find_element("#output")

        data_init = output.text
        log.info("Initial annotation: {0}".format(data_init))
        annotations_init = json.loads(data_init)
        assert len(annotations_init["data"]) == 1
        assert annotations_init["data"][0]["id"] == "TestID"

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))

        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert annotations_updated["timestamp"] > annotations_init["timestamp"]
        assert len(annotations_updated["data"]) == 2
        anno_new = next(
            anno for anno in annotations_updated["data"] if anno["id"] != "TestID"
        )
        assert (
            abs(anno_new["mark"]["width"] - anno_new["mark"]["height"])
            / max(1, abs(anno_new["mark"]["width"]))
            < 0.05
        )

    def test_minimal_change_size(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the minimal app and change the size of an annotation box.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        output: WebElement = dash_duo.find_element("#output")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))
        click_relatively(dash_duo, annotator, pos=(150, 150))
        wait_for_dcc_loading(dash_duo, "#output")
        drag_and_drop(dash_duo, annotator, pos_start=(300, 300), pos_end=(500, 500))

        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 2
        anno_new = next(
            anno for anno in annotations_updated["data"] if anno["id"] != "TestID"
        )
        assert (
            abs(anno_new["mark"]["width"] - anno_new["mark"]["height"])
            / max(1, abs(anno_new["mark"]["width"]))
            < 0.05
        )

    def test_minimal_change_annotation(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the minimal app and change an annotation box.

        The events of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        output: WebElement = dash_duo.find_element("#output")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))
        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation (add): {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 2

        click_relatively(dash_duo, annotator, pos=(150, 150))
        get_input_box(dash_duo, annotator).send_keys("New Annotation")
        click_relatively(dash_duo, annotator, pos=(50, 50))

        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation (name): {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 2
        anno_new = next(
            anno for anno in annotations_updated["data"] if anno["id"] != "TestID"
        )
        assert anno_new["comment"] == "New Annotation"

        click_relatively(dash_duo, annotator, pos=(150, 150))
        get_input_del_button(dash_duo, annotator).click()

        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation (delete): {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 1
        assert annotations_updated["data"][0]["id"] == "TestID"

    def test_minimal_switch_between(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the minimal app and switch between two input boxes.

        The events of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        output: WebElement = dash_duo.find_element("#output")

        # Create boxes
        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))
        drag_and_drop(dash_duo, annotator, pos_start=(400, 100), pos_end=(600, 300))
        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation (add): {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 3

        # Add text comments
        click_relatively(dash_duo, annotator, pos=(150, 150))
        get_input_box(dash_duo, annotator).send_keys("Box1")
        click_relatively(dash_duo, annotator, pos=(50, 50))

        click_relatively(dash_duo, annotator, pos=(450, 150))
        get_input_box(dash_duo, annotator).send_keys("Box2")
        click_relatively(dash_duo, annotator, pos=(50, 50))

        # Check the switching between boxes
        click_relatively(dash_duo, annotator, pos=(150, 150))
        box_text = get_input_box(dash_duo, annotator).get_attribute("value")
        assert box_text == "Box1"
        log.info("Validate the comment text: {0}".format(box_text))
        click_relatively(dash_duo, annotator, pos=(450, 150))
        box_text = get_input_box(dash_duo, annotator).get_attribute("value")
        assert box_text == "Box2"
        log.info("Validate the comment text: {0}".format(box_text))
        click_relatively(dash_duo, annotator, pos=(150, 150))
        box_text = get_input_box(dash_duo, annotator).get_attribute("value")
        assert box_text == "Box1"
        log.info("Validate the comment text: {0}".format(box_text))



# -*- coding: UTF-8 -*-
"""
Size Control
============
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
The tests for the `size_control.py` application. These tests will run a browser emulator
powered by `selenium` and `dash.testing`. The basic functionalities of the demo
will be checked one by one.
"""

import json
import logging

try:
    from typing import Generator
except ImportError:
    from collections.abc import Generator

import pytest

import dash
from dash.testing.application_runners import import_app
from dash.testing.composite import DashComposite
from selenium.webdriver.remote.webelement import WebElement

from .utils import wait_for_dcc_loading
from .actions import drag_and_drop, click_relatively


__all__ = ("TestSizeControl",)


@pytest.mark.with_dash
class TestSizeControl:
    """Test the rendering the Dash app: Size Control"""

    @pytest.fixture(scope="class")
    def dash_app(self) -> Generator[dash.Dash, None, None]:
        log = logging.getLogger("dash_picture_annotation.test")
        log.info("Initialize the Dash app.")
        app = import_app("examples.size_control")
        yield app
        log.info("Remove the Dash app.")
        del app

    def test_sizecontrol_create_new_annotation(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the size_control app and create a new annotation box.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        output: WebElement = dash_duo.find_element("#output")

        data_init = output.text
        log.info("Initial annotation: {0}".format(data_init))
        assert data_init.casefold().strip() == ""

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))
        drag_and_drop(dash_duo, annotator, pos_start=(400, 100), pos_end=(600, 300))

        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 2
        for anno in annotations_updated["data"]:
            assert (
                abs(anno["mark"]["width"] - anno["mark"]["height"])
                / max(1, abs(anno["mark"]["width"]))
                < 0.05
            )

        drag_and_drop(dash_duo, annotator, pos_start=(100, 400), pos_end=(300, 410))
        click_relatively(dash_duo, annotator, (50, 50))
        drag_and_drop(dash_duo, annotator, pos_start=(400, 400), pos_end=(600, 410))
        click_relatively(dash_duo, annotator, (50, 50))
        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 2
        log.info("Confirm that creating a small annotation box will be disallowed.")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 400), pos_end=(300, 600))
        drag_and_drop(dash_duo, annotator, pos_start=(400, 400), pos_end=(600, 600))
        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 4
        log.info("Allowed to create annotations as long as they are big enough.")

        click_relatively(dash_duo, annotator, (450, 450))
        drag_and_drop(dash_duo, annotator, pos_start=(600, 600), pos_end=(600, 410))
        click_relatively(dash_duo, annotator, (50, 50))
        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 3
        log.info(
            "Confirm the auto-delete caused by resizing the annotation box to a "
            "small size."
        )



# -*- coding: UTF-8 -*-
"""
Dash Picture Annotation
=======================

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Dash porting version of the React project React Picture Annotation. Provide a simple
annotation window for a single picture.

Thanks to
---------
Kunduin/react-picture-annotation

https://github.com/Kunduin/react-picture-annotation
"""

import os as _os
import sys as _sys
import json

import dash as _dash

# noinspection PyUnresolvedReferences
from ._imports_ import DashPictureAnnotation
from ._imports_ import __all__ as __import_all__

from . import typehints
from . import utilities

from .typehints import (
    AnnoMark,
    AnnoItem,
    Annotations,
    AnnoStyle,
    DashSelectOptionItem,
    Size,
    Scale,
    NSAnnoItem,
    NSAnnotations,
    is_sequence_of,
    is_anno_mark,
    is_anno_item,
    is_annotations,
    is_dash_select_option_item,
)
from .utilities import (
    get_data_item,
    get_data_item_with_default,
    get_data_items,
    get_data_items_by_regex,
    get_all_ids,
    get_all_comments,
    compare_anno_marks,
    sanitize_data_item,
    sanitize_data,
    sanitize_scale,
)

__all__ = (
    "typehints",
    "utilities",
    "AnnoMark",
    "AnnoItem",
    "Annotations",
    "AnnoStyle",
    "DashSelectOptionItem",
    "Size",
    "Scale",
    "NSAnnoItem",
    "NSAnnotations",
    "is_sequence_of",
    "is_anno_mark",
    "is_anno_item",
    "is_annotations",
    "is_dash_select_option_item",
    "get_data_item",
    "get_data_item_with_default",
    "get_data_items",
    "get_data_items_by_regex",
    "get_all_ids",
    "get_all_comments",
    "compare_anno_marks",
    "sanitize_data_item",
    "sanitize_data",
    "sanitize_scale",
    "DashPictureAnnotation",
)

if not hasattr(_dash, "__plotly_dash") and not hasattr(_dash, "development"):
    print(
        "Dash was not successfully imported. "
        "Make sure you don't have a file "
        'named \n"dash.py" in your current directory.',
        file=_sys.stderr,
    )
    _sys.exit(1)

_basepath = _os.path.dirname(__file__)
_filepath = _os.path.abspath(_os.path.join(_basepath, "package-info.json"))
with open(_filepath) as f:
    package = json.load(f)

package_name = package["name"].replace(" ", "_").replace("-", "_")
__version__ = package["version"]

_current_path = _os.path.dirname(_os.path.abspath(__file__))

_this_module = _sys.modules[__name__]

async_resources = [
    "DashPictureAnnotation",
]

_js_dist = []

_js_dist.extend(
    [
        {
            "relative_package_path": "async-{}.js".format(async_resource),
            "external_url": ("https://unpkg.com/{0}@{2}" "/{1}/async-{3}.js").format(
                package_name, __name__, __version__, async_resource
            ),
            "namespace": package_name,
            "async": True,
        }
        for async_resource in async_resources
    ]
)

# TODO: Figure out if unpkg link works
_js_dist.extend(
    [
        {
            "relative_package_path": "async-{}.js.map".format(async_resource),
            "external_url": (
                "https://unpkg.com/{0}@{2}" "/{1}/async-{3}.js.map"
            ).format(package_name, __name__, __version__, async_resource),
            "namespace": package_name,
            "dynamic": True,
        }
        for async_resource in async_resources
    ]
)

_js_dist.extend(
    [
        {
            "relative_package_path": "dash_picture_annotation.min.js",
            "namespace": package_name,
        },
        {
            "relative_package_path": "dash_picture_annotation.min.js.map",
            "namespace": package_name,
            "dynamic": True,
        },
    ]
)

_css_dist = []


for _component in __import_all__:
    setattr(locals()[_component], "_js_dist", _js_dist)
    setattr(locals()[_component], "_css_dist", _css_dist)



# -*- coding: UTF-8 -*-
"""
Utilities
=========
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Utilities used for managing the data of the annotation results.
"""

import re
import uuid
import hashlib
import datetime
import collections.abc

from typing import Union, Optional, TypeVar

try:
    from typing import Sequence, AbstractSet
    from typing import List, Tuple, Set, FrozenSet
except ImportError:
    from collections.abc import Sequence, Set as AbstractSet
    from builtins import (
        list as List,
        tuple as Tuple,
        set as Set,
        frozenset as FrozenSet,
    )

from typing_extensions import Literal

from . import typehints as th


T = TypeVar("T")

__all__ = (
    "get_data_item",
    "get_data_item_with_default",
    "get_data_items",
    "get_data_items_by_regex",
    "get_all_ids",
    "get_all_comments",
    "compare_anno_marks",
    "sanitize_data_item",
    "sanitize_data",
    "sanitize_scale",
)


def get_data_item(
    data: Union[th.Annotations, Sequence[th.AnnoItem]], id: str
) -> th.AnnoItem:
    """Get an annotation item by its unique ID.

    Note that this method will not validate the format of `data`. If the item cannot
    be found, will raise a `KeyError`.

    Arguments
    ---------
    data: `Annotations | [AnnoItem]`
        The annotation data that is queried.

    id: `str`
        The ID of the annotation item to be found.

    Returns
    -------
    #1: `AnnoItem`
        The queried value. If not found, raise an `KeyError`.
    """
    val = get_data_item_with_default(data, id)
    if val is None:
        raise KeyError('The ID "{0}" is not found in the data.'.format(id))
    return val


def get_data_item_with_default(
    data: Union[th.Annotations, Sequence[th.AnnoItem]], id: str, default: T = None
) -> Union[th.AnnoItem, T]:
    """Get an annotation item by its unique ID with a default value provided.

    Note that this method will not validate the format of `data`. If the item cannot
    be found, will return the default value.

    Arguments
    ---------
    data: `Annotations | [AnnoItem]`
        The annotation data that is queried.

    id: `str`
        The ID of the annotation item to be found.

    default: `T`
        If the ID is not found in the data, return this value.

    Returns
    -------
    #1: `AnnoItem | T`
        The queried value or the default value given by this method.
    """
    if isinstance(data, collections.abc.Mapping):
        data = data["data"]
    if not isinstance(data, collections.abc.Sequence):
        raise TypeError(
            "Cannot iterate the given `data` which type is: "
            "{0}".format(data.__class__.__name__)
        )
    for ditem in data:
        if not isinstance(ditem, collections.abc.Mapping):
            continue
        if "id" not in ditem:
            continue
        val_id = ditem["id"]
        if not isinstance(val_id, str):
            continue
        if val_id == id:
            return ditem
    return default


def get_data_items(
    data: Union[th.Annotations, Sequence[th.AnnoItem]],
    comment: Optional[str] = None,
    n: Optional[int] = None,
) -> Tuple[th.AnnoItem, ...]:
    """Get one or several annotation items by exactly matching the comment.

    Note that this method will not validate the format of each returned annotation
    items. Not well-formatted items will STILL be returned.

    Arguments
    ---------
    data: `Annotations | [AnnoItem]`
        The annotation data that is queried.

    comment: `str | None`
        The comment of the annoatations to be selected. This value will be exactly
        matched in the data, which means that it is space- and case-sensitive.

        If using `None`, will return the annotations without comments.

    n: `int | None`
        Limit the maximal number of the items to be found. If not provided, will try
        to return all annotation items with a specified `comment`.

        Specifying a non-positive value will cause this number to be ignored.

    Returns
    -------
    #1: `[AnnoItem]`
        The located annotation items. Modifying the content of these items will
        cause the `data` to be modified.
    """
    if isinstance(data, collections.abc.Mapping):
        data = data["data"]
    if not isinstance(data, collections.abc.Sequence):
        raise TypeError(
            "Cannot iterate the given `data` which type is: "
            "{0}".format(data.__class__.__name__)
        )
    res: List[th.AnnoItem] = list()
    n_has = 0
    if isinstance(n, int) and n <= 0:
        n = None
    for ditem in data:
        if not isinstance(ditem, collections.abc.Mapping):
            continue
        _comment = ditem.get("comment")
        if _comment == comment:
            res.append(ditem)
            n_has += 1
        if n is not None and n_has >= n:
            break
    return tuple(res)


def get_data_items_by_regex(
    data: Union[th.Annotations, Sequence[th.AnnoItem]],
    comment_regex: Optional[Union[str, re.Pattern]] = None,
    n: Optional[int] = None,
) -> Tuple[th.AnnoItem, ...]:
    """Get one or several annotation items by matching the comment with a regular
    expression.

    The regex is done by `re.match`.

    Note that this method will not validate the format of each returned annotation
    items. Not well-formatted items will STILL be returned.

    Arguments
    ---------
    data: `Annotations | [AnnoItem]`
        The annotation data that is queried.

    comment_regex: `str | re.Pattern | None`
        The regular expression used for matching the comment of the annoatations.
        All comments matching the regex can be returned.

        If using `None`, will return the annotations without comments.

    n: `int | None`
        Limit the maximal number of the items to be found. If not provided, will try
        to return all annotation items with the `comment` matched.

        Specifying a non-positive value will cause this number to be ignored.

    Returns
    -------
    #1: `[AnnoItem]`
        The located annotation items. Modifying the content of these items will
        cause the `data` to be modified.
    """
    if isinstance(data, collections.abc.Mapping):
        data = data["data"]
    if not isinstance(data, collections.abc.Sequence):
        raise TypeError(
            "Cannot iterate the given `data` which type is: "
            "{0}".format(data.__class__.__name__)
        )
    res: List[th.AnnoItem] = list()
    if isinstance(comment_regex, str):
        comment_regex = re.compile(comment_regex)
    if comment_regex is not None and (not isinstance(comment_regex, re.Pattern)):
        raise TypeError(
            'The argument "comment_regex" is not a valid regular expression '
            "pattern: {0}".format(comment_regex)
        )
    n_has = 0
    if isinstance(n, int) and n <= 0:
        n = None
    for ditem in data:
        if not isinstance(ditem, collections.abc.Mapping):
            continue
        _comment = ditem.get("comment")
        if (comment_regex is None and _comment is None) or (
            comment_regex is not None
            and _comment is not None
            and re.match(comment_regex, _comment) is not None
        ):
            res.append(ditem)
            n_has += 1
        if n is not None and n_has >= n:
            break
    return tuple(res)


def get_all_ids(data: Union[th.Annotations, Sequence[th.AnnoItem]]) -> Tuple[str, ...]:
    """Get the list of all IDs in the annotation data. The returne IDs are ordered
    as the order of the items in `data`.

    Arguments
    ---------
    data: `Annotations | [AnnoItem]`
        The annotation data that is queried.

    Returns
    -------
    #1: `[str]`
        The ordered IDs of the annotation items in the data.
    """
    if isinstance(data, collections.abc.Mapping):
        data = data["data"]
    if not isinstance(data, collections.abc.Sequence):
        raise TypeError(
            "Cannot iterate the given `data` which type is: "
            "{0}".format(data.__class__.__name__)
        )
    return tuple(
        ditem["id"]
        for ditem in data
        if isinstance(ditem, collections.abc.Mapping)
        and isinstance(ditem.get("id"), str)
    )


def get_all_comments(
    data: Union[th.Annotations, Sequence[th.AnnoItem]]
) -> FrozenSet[str]:
    """Get the set of all comments in the annotation data. The returned comments are
    unordered.

    Arguments
    ---------
    data: `Annotations | [AnnoItem]`
        The annotation data that is queried.

    Returns
    -------
    #1: `{str}`
        The unordered set of comments appearing in the annotation data. If the data
        does not contain any comment, will return an empty set.
    """
    if isinstance(data, collections.abc.Mapping):
        data = data["data"]
    if not isinstance(data, collections.abc.Sequence):
        raise TypeError(
            "Cannot iterate the given `data` which type is: "
            "{0}".format(data.__class__.__name__)
        )
    res: Set[str] = set()
    for ditem in data:
        if not isinstance(ditem, collections.abc.Mapping):
            continue
        _comment = ditem.get("comment")
        if isinstance(_comment, str):
            res.add(_comment)
    return frozenset(res)


def compare_anno_marks(
    anno_item_1: Union[th.AnnoItem, th.AnnoMark],
    anno_item_2: Union[th.AnnoItem, th.AnnoMark],
    tolearance: float = 0.1,
) -> bool:
    """Compare whether two annotation marks are equivalent or not.

    The comparison is only performed on the positions. The ID and the comment of the
    annotation items will not be compared.

    Arguments
    ---------
    anno_item_1: `AnnoItem | AnnoMark`
    anno_item_2: `AnnoItem | AnnoMark`
        The annotation items to be compared.

    tolearance: `float`
        The tolerance used for checking whether positions are the same or not.

    Returns
    -------
    #1: `bool`
        A flag. Returns `True` if the two items are in the same position and share the
        same size.

        Will return `False` if any of them is not an annotation item or mark.
    """
    if not (
        isinstance(anno_item_1, collections.abc.Mapping)
        and isinstance(anno_item_2, collections.abc.Mapping)
    ):
        return False
    if "mark" in anno_item_1:
        anno_item_1 = anno_item_1["mark"]
    if "mark" in anno_item_2:
        anno_item_2 = anno_item_2["mark"]
    if not (th.is_anno_mark(anno_item_1) and th.is_anno_mark(anno_item_2)):
        return False
    if abs(abs(anno_item_1["width"]) - abs(anno_item_2["width"])) > tolearance:
        return False
    if abs(abs(anno_item_1["height"]) - abs(anno_item_2["height"])) > tolearance:
        return False
    if (
        abs(
            (
                anno_item_1["x"]
                if anno_item_1["width"] > 0
                else (anno_item_1["x"] + anno_item_1["width"])
            )
            - (
                anno_item_2["x"]
                if anno_item_2["width"] > 0
                else (anno_item_2["x"] + anno_item_2["width"])
            )
        )
        > tolearance
    ):
        return False
    if (
        abs(
            (
                anno_item_1["y"]
                if anno_item_1["height"] > 0
                else (anno_item_1["y"] + anno_item_1["height"])
            )
            - (
                anno_item_2["y"]
                if anno_item_2["height"] > 0
                else (anno_item_2["y"] + anno_item_2["height"])
            )
        )
        > tolearance
    ):
        return False
    return True


def _generate_id(n: int = 3) -> str:
    """Generate the random ID. (private method)

    Arguments
    ---------
    n: `int`
        The number of characters of the ID.

    Returns
    -------
    #1: `str`
        The randomly generated ID.
    """
    hash_code = hashlib.shake_256()
    hash_code.update(uuid.uuid4().bytes)
    return hash_code.hexdigest(n)


def sanitize_data_item(data_item: th.NSAnnoItem) -> Optional[th.AnnoItem]:
    """Perform the full sanitization on an annotation item.

    The sanitization will ensure that:
    1. The result is a dictionary typed by `AnnoItem`.
    2. A `None` value comment will be removed.
    3. The position of the annotation mark will always have positive width and height.
       (Negative values means that the starting location is reversed.)
    4. If no ID is specified, will return a randomly generated 6-character ID.

    Arguments
    ---------
    data_item: `AnnoItem | AnnoMark | {str: Any}`
        The annotation item data to be sanitized. If the `data_item` is just a mark,
        will add ID for it.

    Returns
    -------
    #1: `AnnoItem | None`
        The sanitized copy of the data item. Will return `None` if the given data
        cannot be sanitized.
    """
    if not isinstance(data_item, collections.abc.Mapping):
        return None
    if "mark" not in data_item:
        if th.is_anno_mark(data_item):
            data_item = th.AnnoItem(id=_generate_id(), mark=data_item)
        else:
            return None
    else:
        if not th.is_anno_mark(data_item["mark"]):
            return None
    sanitized_id = data_item.get("id")
    comment = data_item.get("comment")
    if not sanitized_id:
        sanitized_id = _generate_id()
    data_item = th.AnnoItem(id=sanitized_id, mark=data_item["mark"].copy())
    if isinstance(comment, str):
        data_item["comment"] = comment
    mark = data_item["mark"]
    width = mark["width"]
    height = mark["height"]
    if width < 0:
        mark["width"] = -width
        mark["x"] = mark["x"] + width
    if height < 0:
        mark["height"] = -height
        mark["y"] = mark["y"] + height
    return data_item


def _deduplicate_id_by_add(seen_ids: AbstractSet[str], anno_id: str) -> str:
    """Ensure that the given `anno_id` is not duplicated by the "add" rule.
    (private method)

    Arguments
    ---------
    seen_ids: `{str}`
        The collection of seen IDs.

    anno_id: `str`
        The new ID to be checked.

    Returns
    -------
    #1: `str`
        If `anno_id` is not in `seen_ids`, return it as it is. Otherwise, return a
        modified ID that is not in `seen_ids`.
    """
    _anno_id = anno_id
    while _anno_id in seen_ids:
        _anno_id = anno_id + _generate_id()
    return _anno_id


def sanitize_data(
    data: Union[th.NSAnnotations, Sequence[th.NSAnnoItem]],
    deduplicate: Literal["add", "drop"] = "drop",
) -> th.Annotations:
    """Perform the full sanitization on the annotation data.

    The sanitization will ensure that:
    1. The current timestamp is the time when this method is used.
    2. All items in `data` are dictionaries typed by `AnnoItem`.
    3. Any item with a `None` comment will be sanitized as an item without the comment.
    4. The position of the annotation mark will always have positive width and height.
       (Negative values means that the starting location is reversed.)
    4. Annotation items will be deduplicated, i.e. the IDs will be sanitized.

    This method can be used when the annotations need to be saved as a file. It may
    take time to run, so it may not be suitable to sanitize the annotations in the
    real time.

    Arguments
    ---------
    data: `Annotations | [AnnoItem | {str: Any}]`
        The annotation data that will be sanitized. Note that this method will not
        change the input data.

    deduplicate: `"add" | "drop"`
        The deduplicate method for the annotation IDs. `"add"` means that preserving
        the duplicated ID by adding a postfix. `"drop"` means that dropping all
        annotation items with duplicated IDs after the first found item.

    Returns
    -------
    #1: `Annotations`
        The sanitized copy of the data.
    """
    if isinstance(data, collections.abc.Mapping):
        data = data["data"]
    if not isinstance(data, collections.abc.Sequence):
        return th.Annotations(
            timestamp=int(datetime.datetime.now().timestamp() * 1000), data=list()
        )
    new_data: List[th.AnnoItem] = list()
    seen_ids: Set[str] = set()
    for ditem in data:
        new_anno = sanitize_data_item(ditem)
        if new_anno is None:
            continue
        anno_id = new_anno.get("id")
        if not anno_id:
            anno_id = _generate_id()
        if anno_id in seen_ids:
            if deduplicate == "add":
                anno_id = _deduplicate_id_by_add(seen_ids, anno_id)
            else:
                continue
        new_anno["id"] = anno_id
        seen_ids.add(anno_id)
        new_data.append(new_anno)
    return th.Annotations(
        timestamp=int(datetime.datetime.now().timestamp() * 1000), data=new_data
    )


def sanitize_scale(
    scale: Union[float, th.Mapping[str, th.Any]],
    offset_x: Optional[float] = None,
    offset_y: Optional[float] = None,
) -> th.Scale:
    """Perform the sanitization on the annotated image scaling factor.

    The sanitization will ensure that:
    1. A dictionary containing at least the scaling ratio will be returned.
    2. Optional offsets can be configured.
    3. A timestamp will be attached to the sanitized configuration. It ensures that the
       scaling event will be always triggered even if the configuration does not
       change.

    Arguments
    ---------
    scale: `float | Mapping[str, Any]`
        The scaling factor value (float) or a full scaling configuration dictionary.

    offset_x: `float`
        The relative offset ratio along the X axis. This value will be added to the
        returned value only when it is configured and `"offset_x"` is not configured
        in `scale`.

    offset_y: `float`
        The relative offset ratio along the Y axis. This value will be added to the
        returned value only when it is configured and `"offset_y"` is not configured
        in `scale`.

    Returns
    -------
    #1: `Scale`
        The santized scaling factor with a newly configured timestamp.
    """
    if isinstance(scale, collections.abc.Mapping):
        res = th.Scale(scale=(float(scale["scale"]) if "scale" in scale else 1.0))
        if "offset_x" in scale:
            res["offset_x"] = float(scale["offset_x"])
        if "offset_y" in scale:
            res["offset_y"] = float(scale["offset_y"])
    else:
        res = th.Scale(scale=float(scale))
    if "offset_x" not in res and isinstance(offset_x, float):
        res["offset_x"] = offset_x
    if "offset_y" not in res and isinstance(offset_y, float):
        res["offset_y"] = offset_y
    res["timestamp"] = int(datetime.datetime.now().timestamp() * 1000)
    return res



# -*- coding: UTF-8 -*-
"""
Options
=======
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
The tests for the `options.py` application. These tests will run a browser emulator
powered by `selenium` and `dash.testing`. The basic functionalities of the demo
will be checked one by one.
"""

import json
import logging

try:
    from typing import Generator
except ImportError:
    from collections.abc import Generator

import pytest

import dash
from dash.testing.application_runners import import_app
from dash.testing.composite import DashComposite
from selenium.webdriver.remote.webelement import WebElement

from .utils import wait_for_the_attribute_value_to_equal, wait_for_dcc_loading, wait_for
from .actions import (
    drag_and_drop,
    click_relatively,
    get_select_expand,
    get_select_del_button,
    get_select_control_points,
    get_select_item_by_label,
    get_select_current_item,
)


__all__ = ("TestOptions",)


@pytest.mark.with_dash
class TestOptions:
    """Test the rendering the Dash app: Options"""

    @pytest.fixture(scope="class")
    def dash_app(self) -> Generator[dash.Dash, None, None]:
        log = logging.getLogger("dash_picture_annotation.test")
        log.info("Initialize the Dash app.")
        app = import_app("examples.options")
        yield app
        log.info("Remove the Dash app.")
        del app

    def test_options_toggle_options(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the options app and toggle the option box.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_options: WebElement = dash_duo.find_element("#btn-options")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))
        click_relatively(dash_duo, annotator, (150, 150))

        dash_duo.wait_for_element_by_css_selector(
            "#annotator > .rp-stage > .rp-selected-input > "
            "div.rp-default-input-section > input"
        )
        log.info("Current annotation is controlled by <input>.")

        btn_options.click()
        dash_duo.wait_for_element_by_css_selector(
            "#annotator > .rp-stage > .rp-selected-input > div[class^='dpa-dropdown']",
        )
        log.info("Current annotation is controlled by <Select>.")

        btn_options.click()
        dash_duo.wait_for_element_by_css_selector(
            "#annotator > .rp-stage > .rp-selected-input > "
            "div.rp-default-input-section > input"
        )
        log.info("Current annotation is controlled by <input>.")

    def test_options_toggle_option_clearable(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the options app and toggle the option box.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_options: WebElement = dash_duo.find_element("#btn-options")
        btn_optclear: WebElement = dash_duo.find_element("#btn-optclear")
        output: WebElement = dash_duo.find_element("#output")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))
        click_relatively(dash_duo, annotator, (150, 150))

        dash_duo.wait_for_element_by_css_selector(
            "#annotator > .rp-stage > .rp-selected-input > "
            "div.rp-default-input-section > input"
        )
        log.info("Current annotation is controlled by <input>.")

        btn_options.click()
        dash_duo.wait_for_element_by_css_selector(
            "#annotator > .rp-stage > .rp-selected-input > div[class^='dpa-dropdown']",
        )
        log.info("Current annotation is controlled by <Select>.")

        # Open the menu
        element = get_select_expand(dash_duo, annotator)
        element.click()

        # Find the first item
        element = get_select_item_by_label(dash_duo, annotator, item_label="Type 1")
        element.click()

        # Validate the currently selected value.
        element = get_select_current_item(dash_duo, annotator, expect_to_be="item")
        wait_for(dash_duo, lambda: (element.text == "Type 1"))
        log.info("Currently selected option: {0}".format(element.text))

        # Validate that it is clearable now or not.
        elements = get_select_control_points(dash_duo, annotator)
        assert len(elements) == 2
        log.info("Currently selected option is clerable.")

        btn_optclear.click()
        elements = get_select_control_points(dash_duo, annotator)
        assert len(elements) == 1
        log.info("Currently selected option is not clerable.")

        btn_optclear.click()
        elements = get_select_control_points(dash_duo, annotator)
        assert len(elements) == 2
        log.info("Currently selected option is clerable.")

        click_relatively(dash_duo, annotator, (50, 50))
        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 1
        assert annotations_updated["data"][0]["comment"] == "type-1"

        # Remove the comment
        click_relatively(dash_duo, annotator, (150, 150))
        elements = get_select_control_points(dash_duo, annotator)
        assert len(elements) == 2
        elements[0].click()
        get_select_current_item(dash_duo, annotator, expect_to_be="placeholder")

        click_relatively(dash_duo, annotator, (50, 50))
        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 1
        assert not annotations_updated["data"][0].get("comment")

    def test_options_placeholders(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the options app and validate the customized placeholders.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_options: WebElement = dash_duo.find_element("#btn-options")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))
        click_relatively(dash_duo, annotator, (150, 150))

        wait_for_the_attribute_value_to_equal(
            dash_duo,
            "#annotator > .rp-stage > .rp-selected-input > "
            "div.rp-default-input-section > input",
            attribute="placeholder",
            value="Using free mode...",
        )
        log.info("Confirm the placeholder of <input>.")

        btn_options.click()
        element = get_select_current_item(
            dash_duo, annotator, expect_to_be="placeholder"
        )
        assert element.text == "Using selection mode..."
        log.info("Confirm the placeholder of <Select>.")

    def test_options_delete_annotation(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the options app and remove the annotation box by the button of selector.

        The events of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_options: WebElement = dash_duo.find_element("#btn-options")
        output: WebElement = dash_duo.find_element("#output")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))

        click_relatively(dash_duo, annotator, (150, 150))
        btn_options.click()
        dash_duo.wait_for_element_by_css_selector(
            "#annotator > .rp-stage > .rp-selected-input > div[class^='dpa-dropdown']",
        )
        log.info("Current annotation is controlled by <Select>.")

        click_relatively(dash_duo, annotator, (50, 50))
        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert len(annotations_updated["data"]) == 1

        click_relatively(dash_duo, annotator, (150, 150))
        element = get_select_del_button(dash_duo, annotator)
        element.click()

        wait_for_dcc_loading(dash_duo, "#output")
        data_updated = output.text
        log.info("Updated annotation: {0}".format(data_updated))
        annotations_updated = json.loads(data_updated)
        assert not annotations_updated.get("data")



# -*- coding: UTF-8 -*-
"""
Colors
======
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
The tests for the `colors.py` application. These tests will run a browser emulator
powered by `selenium` and `dash.testing`. The basic functionalities of the demo
will be checked one by one.
"""

import logging

try:
    from typing import Generator
except ImportError:
    from collections.abc import Generator

import pytest

import dash
from dash.testing.application_runners import import_app
from dash.testing.composite import DashComposite
from selenium.webdriver.remote.webelement import WebElement

from .utils import wait_for_dcc_loading, asset_folder
from .imgutils import decode_base64_image, compare_images_with_alpha, ImageLoader
from .actions import get_canvas_image


__all__ = ("TestColors",)


@pytest.mark.with_dash
class TestColors:
    """Test the rendering the Dash app: Colors"""

    @pytest.fixture(scope="class")
    def dash_app(self) -> Generator[dash.Dash, None, None]:
        log = logging.getLogger("dash_picture_annotation.test")
        log.info("Initialize the Dash app.")
        app = import_app("examples.colors")
        yield app
        log.info("Remove the Dash app.")
        del app

    @pytest.fixture(scope="class")
    def ref_images(self) -> Generator[ImageLoader, None, None]:
        """Fixture of the testing images."""
        log = logging.getLogger("dash_picture_annotation.test")
        log.info("Use the reference images.")
        yield ImageLoader(root=asset_folder())

    @staticmethod
    def get_check_function_for_current_shape_ssim(
        dash_duo: DashComposite, ref_images: ImageLoader, annotator: WebElement
    ):
        """Get the check function for the ssim of the current shape.

        This factory method will be used to create the shape validator in the tests.
        It is internally used.

        Arguments
        ---------
        dash_duo: `DashComposite`
        ref_images: `ref_images`
            Fixtures used in the test.

        annotator: `WebElement`
            The annotator box gotten by
            ``` python
            dash_duo.find_element("#annotator")
            ```

        Returns
        -------
        #1: `(name: str) -> float`
            A validator function. It accepts the name of the figure that is used for
            validating the results. If the ssim score is lower than 0.9, throw an
            error. If the check succeeds, return the ssim score.
        """

        def check_function(name: str, validate: bool = True) -> float:
            """Check function for the current shapes.

            Arguments
            ---------
            name: `str`
                The name of the reference image used for performing the check.

            validate: `bool`
                A flag. If specified, will validate the score by `> 0.9`.

            Returns
            -------
            #1: `float`
                If the check succeeds, return the ssim score. Otherwise, throw an
                error.
            """
            wait_for_dcc_loading(dash_duo, "#annotator")
            img = decode_base64_image(
                get_canvas_image(dash_duo, annotator, canvas_selector=".rp-shapes"),
                use_rgba=True,
            )
            ssim = compare_images_with_alpha(img, ref_images[name], fine_mode=True)
            if validate:
                assert ssim > 0.9
            return ssim

        return check_function

    def test_colors_set_color(
        self, dash_duo: DashComposite, dash_app: dash.Dash, ref_images: ImageLoader
    ) -> None:
        """Test the colors app and change the default color of annotations.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        wait_for_dcc_loading(dash_duo, "#annotator")

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_default: WebElement = dash_duo.find_element("#btn-default")
        btn_specific: WebElement = dash_duo.find_element("#btn-specific")

        checker = self.get_check_function_for_current_shape_ssim(
            dash_duo, ref_images, annotator
        )

        wait_for_dcc_loading(dash_duo, "#annotator")
        ssim = checker("ref_color_init")
        log.info("SSIM of the initial annotations: {0:.3g}".format(ssim))

        btn_default.click()
        wait_for_dcc_loading(dash_duo, "#annotator")
        ssim = checker("ref_color_change_default")
        log.info("SSIM of changing the default color: {0:.3g}".format(ssim))

        btn_specific.click()
        wait_for_dcc_loading(dash_duo, "#annotator")
        ssim = checker("ref_color_specific")
        log.info("SSIM of specifying colors of some types: {0:.3g}".format(ssim))

        btn_specific.click()
        wait_for_dcc_loading(dash_duo, "#annotator")
        ssim = checker("ref_color_change_default")
        log.info("SSIM of reverting the specific colors: {0:.3g}".format(ssim))

        btn_default.click()
        wait_for_dcc_loading(dash_duo, "#annotator")
        ssim = checker("ref_color_init")
        log.info("SSIM of reverting the default color: {0:.3g}".format(ssim))

    def test_colors_dynamic_color(
        self, dash_duo: DashComposite, dash_app: dash.Dash, ref_images: ImageLoader
    ) -> None:
        """Test the colors app and toggle the dynamic colors.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        wait_for_dcc_loading(dash_duo, "#annotator")

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_default: WebElement = dash_duo.find_element("#btn-default")
        btn_specific: WebElement = dash_duo.find_element("#btn-specific")
        btn_dynamic: WebElement = dash_duo.find_element("#btn-dynamic")

        checker = self.get_check_function_for_current_shape_ssim(
            dash_duo, ref_images, annotator
        )

        wait_for_dcc_loading(dash_duo, "#annotator")
        ssim = checker("ref_color_init")
        log.info("SSIM of the initial annotations: {0:.3g}".format(ssim))

        btn_dynamic.click()
        wait_for_dcc_loading(dash_duo, "#annotator")
        ssim = checker("ref_color_dynamic")
        log.info("SSIM of changing the dynamic color: {0:.3g}".format(ssim))

        btn_default.click()
        wait_for_dcc_loading(dash_duo, "#annotator")
        _ssim = checker("ref_color_dynamic")
        assert abs(ssim - _ssim) < 1e-3
        log.info(
            "Confirm that default color will not override dynamic "
            "color: {0:.3g}".format(_ssim)
        )

        btn_specific.click()
        wait_for_dcc_loading(dash_duo, "#annotator")
        _ssim = checker("ref_color_dynamic", validate=False)
        assert _ssim < ssim
        assert abs(ssim - _ssim) > 1e-3
        log.info(
            "Confirm that specific color will override dynamic color: "
            "{0:.3g}".format(_ssim)
        )
        ssim = checker("ref_color_dynamic_specific")
        assert _ssim < ssim
        assert abs(ssim - _ssim) > 1e-3
        log.info(
            "Confirm that dynamic color with specified color matches the target: "
            "{0:.3g}".format(ssim)
        )

        btn_specific.click()
        btn_dynamic.click()
        btn_default.click()
        wait_for_dcc_loading(dash_duo, "#annotator")
        ssim = checker("ref_color_init")
        log.info("SSIM of reverting the dynamic colors: {0:.3g}".format(ssim))



# -*- coding: UTF-8 -*-
"""
Utilities
=========
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Extra functionalities used for enhancing the tests.
"""

import os
import io
import base64

from typing import Union, Optional

try:
    from typing import Sequence
    from typing import Tuple
except ImportError:
    from collections.abc import Sequence
    from builtins import tuple as Tuple

from typing_extensions import Literal

from PIL import Image, ImageOps, ImageChops
import numpy as np
from skimage.metrics import structural_similarity as ssim


__all__ = (
    "image_rgba_to_rgb",
    "remove_image_margin",
    "decode_base64_image",
    "compare_images",
    "compare_images_with_alpha",
    "ImageLoader",
)


def image_rgba_to_rgb(img: Image.Image, background: str = "black") -> Image.Image:
    """Convert an RGBA image to an RGB image by adding a background color."""
    return Image.alpha_composite(Image.new("RGBA", img.size, background), img).convert(
        "RGB"
    )


def remove_image_margin(img: Image.Image) -> Image.Image:
    """Remove the margin of an image."""
    if img.mode.casefold() == "rgba":
        bounding_box = img.getbbox()
        if bounding_box:
            img = img.crop(bounding_box)
        img_a = ImageOps.invert(image_rgba_to_rgb(img, "white"))
        img_b = image_rgba_to_rgb(img, "black")
    else:
        img_b = img.convert("RGB")
        img_a = ImageOps.invert(img_b)
    bounding_box = ImageChops.lighter(
        ImageChops.subtract(
            img_a, Image.new("RGB", img_a.size, img_a.getpixel((0, 0))), 1.0, -100
        ),
        ImageChops.subtract(
            img_b, Image.new("RGB", img_b.size, img_b.getpixel((0, 0))), 1.0, -100
        ),
    ).getbbox()
    if bounding_box:
        img = img.crop(bounding_box)
    return img


def decode_base64_image(text: str, use_rgba: bool = False) -> Image.Image:
    """Decode the base64 encoded png image to the correct image."""
    img = Image.open(
        io.BytesIO(base64.b64decode(text[len("data:image/png;base64,") :])),
        formats=("png", "webp", "jpeg"),
    )
    img.load()
    img = remove_image_margin(img)
    if img.mode.casefold() == "rgba":
        if not use_rgba:
            img = image_rgba_to_rgb(img, "white")
    if use_rgba:
        if img.mode.casefold() != "rgba":
            img = img.convert("RGBA")
    else:
        img = img.convert("RGB")
    return img


def compare_images(
    image: Image.Image, ref: Image.Image, size: Optional[Tuple[int, int]] = None
) -> float:
    """Compare two images, calculate the similarity bettween then, assuming that the
    images do not contain the alpha channels.

    Arguments
    ---------
    image: `Image`
        The image to be checked.

    ref: `Image`
        The reference image to be used for comparison. Note that the reference image
        will be used for specifying the data range.

    size: `(int, int) | None`
        The resize target applied to the images before the comparison. If not
        specified, will resize the width to be 100.
    """
    if size is None:
        size = (100, max(1, int(round(ref.size[1] * 100 / ref.size[0]))))
    image_data = np.asarray(
        image.convert(mode="RGB").resize(size, Image.Resampling.LANCZOS)
    )
    ref_data = np.asarray(
        ref.convert(mode="RGB").resize(size, Image.Resampling.LANCZOS)
    )
    return float(
        ssim(
            image_data,
            ref_data,
            data_range=ref_data.max() - ref_data.min(),
            channel_axis=2,
        )
    )


def compare_images_with_alpha(
    image: Image.Image,
    ref: Image.Image,
    size: Optional[Tuple[int, int]] = None,
    fine_mode: bool = False,
) -> float:
    """Compare two images, calculate the similarity bettween them, and weight the
    similarity by the alpha channel.

    Arguments
    ---------
    image: `Image`
        The image to be checked.

    ref: `Image`
        The reference image to be used for comparison. Note that the reference image
        will be used for specifying the data range and the alpha channel weight.

    size: `(int, int) | None`
        The resize target applied to the images before the comparison. If not
        specified, will resize the width to be 100.

    fine_mode: `bool`
        A flag. Whether to turn on the fine mode. If specified, will not use the
        binary alpha channel to select and calculate the averaged ssim, but use the
        alpha channel as the weight of the ssim. This configuration will make the
        weight of ssim values more fine-grained.
    """
    if size is None:
        size = (100, max(1, int(round(ref.size[1] * 100 / ref.size[0]))))
    image = image.convert(mode="RGBA").resize(size, Image.Resampling.LANCZOS)
    ref = ref.convert(mode="RGBA").resize(size, Image.Resampling.LANCZOS)

    ref_data = np.asarray(image_rgba_to_rgb(ref, "white"))
    _, ssim_img_white = ssim(
        np.asarray(image_rgba_to_rgb(image, "white")),
        ref_data,
        data_range=ref_data.max() - ref_data.min(),
        channel_axis=2,
        full=True,
    )

    ref_data = np.asarray(image_rgba_to_rgb(ref, "black"))
    _, ssim_img_black = ssim(
        np.asarray(image_rgba_to_rgb(image, "black")),
        ref_data,
        data_range=ref_data.max() - ref_data.min(),
        channel_axis=2,
        full=True,
    )

    val_ssim = np.asarray(np.minimum(ssim_img_black, ssim_img_white))
    if fine_mode:
        mask_alpha = np.asarray(ref.split()[-1])
        mask_alpha = mask_alpha / max(1.0, np.amax(mask_alpha))
        weight = np.sum(mask_alpha) * val_ssim.shape[-1]
        if weight < 1e-3:
            return 1.0
        else:
            return float(
                np.sum(
                    np.reshape(val_ssim, (-1, val_ssim.shape[-1]))
                    * np.expand_dims(mask_alpha.flatten(), axis=-1)
                )
                / weight
            )
    else:
        mask_alpha = np.asarray(ref.split()[-1]) < 100
        return float(
            np.mean(
                np.reshape(val_ssim, (-1, val_ssim.shape[-1]))[mask_alpha.flatten(), :]
            )
        )


class ImageLoader:
    """Loader for images.

    This object is used for dynamically loading images from a specific folder during
    the test.
    """

    def __init__(
        self,
        root: str,
        mode: Literal["RGB", "RGBA"] = "RGBA",
        ext_list: Union[Sequence[str], str] = "webp",
    ) -> None:
        """Initialization.

        Arguments
        ---------
        root: `str`
            The root path of the images to be read.

        mode: `"RGB" | "RGBA"`
            The mode used for saving images.

        ext_list: `[str] | str`
            The list of file name extensions to be searched.
        """
        self.root = str(root)
        if not os.path.isdir(self.root):
            raise FileNotFoundError(
                'The argument "root" specifies a path that is not a folder.'
            )
        self.mode = mode
        self.ext_list = (ext_list,) if isinstance(ext_list, str) else tuple(ext_list)

    def __normalize_path(self, file_path: str) -> str:
        """Normalize the path of a file for ensuring that the path refers to an
        existing file.

        If no valid file can be detected, will return the input `file_path` directly.
        """
        if os.path.isfile(file_path):
            return file_path
        for ext in self.ext_list:
            _file_path = "{0}.{1}".format(file_path, ext)
            if os.path.isfile(_file_path):
                return _file_path
        _file_path = os.path.splitext(file_path)[0]
        if os.path.isfile(_file_path):
            return _file_path
        for ext in self.ext_list:
            _file_path_v2 = "{0}.{1}".format(_file_path, ext)
            if os.path.isfile(_file_path_v2):
                return _file_path_v2
        return file_path

    def __getitem__(self, file_name: str) -> Image.Image:
        """Load a image by calling the `__getitem__` method."""
        file_path = self.__normalize_path(os.path.join(self.root, file_name))
        with open(file_path, "rb") as fobj:
            img = Image.open(fobj, formats=self.ext_list)
            img.load()
        return img

    def __setitem__(self, file_name: str, img: Image.Image) -> None:
        """Save the image as a reference image in the root folder."""
        file_path = os.path.join(self.root, file_name)
        if not os.path.isfile(file_path):
            file_path_base, ext = os.path.splitext(file_name)
            if ext.casefold().strip().strip(".") not in self.ext_list:
                ext = ".{0}".format(self.ext_list[0])
            file_path = file_path_base + ext
        size = (100, max(1, int(round(img.size[1] * 100 / img.size[0]))))
        ext = os.path.splitext(file_path)[-1].casefold().strip().strip(".")
        img.convert(mode=self.mode).resize(size, Image.Resampling.LANCZOS).save(
            os.path.join("tests", "assets", file_path),
            format=ext if ext in self.ext_list else "webp",
            lossless=True,
            exact=True,
        )



# -*- coding: UTF-8 -*-
"""
Options
=======
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

License
-------
MIT License

Description
-----------
The demo for using the selecting box to specify the annotation comments.
"""

import os
import json

from typing import Optional, Any

try:
    from typing import Sequence
except ImportError:
    from collections.abc import Sequence

import dash
from dash import dcc, html
from dash import Output, Input, State


if __name__ == "__main__":
    import sys

    sys.path.append(os.path.dirname(os.path.dirname(__file__)))


import dash_picture_annotation as dpa


app = dash.Dash(
    __name__,
    assets_folder=os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "tests", "assets"
    ),
)

default_data = {"timestamp": 0, "data": []}

default_options = [
    {"value": "type-1", "label": "Type 1"},
    {"value": "type-2", "label": "Type 2"},
]

styles = {
    "mb1": {"marginBottom": "1rem"},
    "mr1": {"marginRight": "1rem"},
}

app.layout = html.Div(
    [
        dcc.Loading(
            dpa.DashPictureAnnotation(
                id="annotator",
                style={"height": "80vh"},
                data=default_data,
                image="/assets/test_image.svg",
                options=None,
                clearable_dropdown=True,
                placeholder_input="Using free mode...",
                placeholder_dropdown="Using selection mode...",
            ),
            delay_show=1000,
        ),
        dcc.Loading(
            html.Div(
                id="output", children=json.dumps(default_data), style=styles["mb1"]
            ),
            delay_show=1000,
        ),
        html.Div(
            children=(
                html.Button(
                    children="Toggle options", id="btn-options", style=styles["mr1"]
                ),
                html.Button(children="Toggle options clearable", id="btn-optclear"),
            ),
            style=styles["mb1"],
        ),
    ]
)


@app.callback(
    Output("output", "children"), Input("annotator", "data"), prevent_initial_call=False
)
def get_annotation(data) -> Optional[str]:
    if data is None:
        return None
    return json.dumps(data)


@app.callback(
    Output("annotator", "options"),
    Output("annotator", "clearable_dropdown"),
    Input("btn-options", "n_clicks"),
    Input("btn-optclear", "n_clicks"),
    State("annotator", "options"),
    State("annotator", "clearable_dropdown"),
    prevent_initial_call=True,
)
def toggle_options(
    n_clicks_options: Optional[int],
    n_clicks_optclear: Optional[int],
    options: Optional[Sequence[Any]],
    clearable_dropdown: Optional[bool],
):
    trigger = dash.ctx.triggered_id
    if trigger and trigger == "btn-options" and n_clicks_options:
        if options:
            return None, dash.no_update
        else:
            return default_options, dash.no_update
    if trigger and trigger == "btn-optclear" and n_clicks_optclear:
        return dash.no_update, not bool(clearable_dropdown)
    return dash.no_update, dash.no_update


if __name__ == "__main__":
    import socket

    def get_ip(method: str = "broadcast") -> str:
        """Detect the IP address of this device."""
        s_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            if method == "broadcast":
                s_socket.connect(("10.255.255.255", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "udp":
                s_socket.connect(("8.8.8.8", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "host":
                ip_value = socket.gethostbyname(socket.gethostname())
            else:
                raise ConnectionError
        except Exception:  # pylint: disable=broad-except
            ip_value = "localhost"
        finally:
            s_socket.close()
        return ip_value

    app.run(host=get_ip(), port="8080", debug=False)



# -*- coding: UTF-8 -*-
"""
Colors
======
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

License
-------
MIT License

Description
-----------
The demo for configuring colors of the annotation boxes.
"""

import os
import json

from typing import Optional

try:
    from typing import Mapping
except ImportError:
    from collections.abc import Mapping

import dash
from dash import dcc, html
from dash import Output, Input, State


if __name__ == "__main__":
    import sys

    sys.path.append(os.path.dirname(os.path.dirname(__file__)))


import dash_picture_annotation as dpa


app = dash.Dash(
    __name__,
    assets_folder=os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "tests", "assets"
    ),
)

with open(
    os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "tests", "data-colors.json"
    ),
    "r",
) as fobj:
    example_data = json.load(fobj)

styles = {
    "mb1": {"marginBottom": "1rem"},
    "mr1": {"marginRight": "1rem"},
}

app.layout = html.Div(
    [
        dcc.Loading(
            dpa.DashPictureAnnotation(
                id="annotator",
                style={"height": "80vh"},
                data=example_data,
                image="/assets/test_image.svg",
                options=None,
                clearable_dropdown=True,
            ),
            delay_show=1000,
        ),
        dcc.Loading(
            html.Div(id="output", children=None, style={"marginBottom": "1rem"}),
            delay_show=1000,
        ),
        html.Div(
            children=(
                html.Button(
                    children="Toggle default color",
                    id="btn-default",
                    style=styles["mr1"],
                ),
                html.Button(
                    children="Toggle specific color",
                    id="btn-specific",
                    style=styles["mr1"],
                ),
                html.Button(
                    children="Toggle dynamic color",
                    id="btn-dynamic",
                    style=styles["mr1"],
                ),
            ),
            style=styles["mb1"],
        ),
    ]
)


@app.callback(
    Output("output", "children"), Input("annotator", "data"), prevent_initial_call=False
)
def get_annotation(data) -> Optional[str]:
    if data is None:
        return None
    return json.dumps(data)


@app.callback(
    Output("annotator", "style_annotation"),
    Input("btn-default", "n_clicks"),
    State("annotator", "style_annotation"),
    prevent_initial_call=True,
)
def toggle_default_color(n_clicks: Optional[int], style_annotation: Optional[str]):
    if n_clicks:
        if not isinstance(style_annotation, str):
            return "#000"
        if style_annotation == "#fff":
            return "#000"
        else:
            return "#fff"
    return dash.no_update


@app.callback(
    Output("annotator", "colors"),
    Input("btn-specific", "n_clicks"),
    State("annotator", "colors"),
    prevent_initial_call=True,
)
def toggle_specific_colors(
    n_clicks: Optional[int], colors: Optional[Mapping[str, str]]
):
    if n_clicks:
        if colors:
            return None
        else:
            return dict(
                zip(
                    (ditem["comment"] for ditem in example_data["data"]),
                    ("hsl(180, 100%, 50%)", "rgb(255, 0, 255)", "#ff0"),
                )
            )
    return dash.no_update


@app.callback(
    Output("annotator", "is_color_dynamic"),
    Input("btn-dynamic", "n_clicks"),
    State("annotator", "is_color_dynamic"),
    prevent_initial_call=True,
)
def toggle_dynamic_color(n_clicks: Optional[int], is_color_dynamic: Optional[bool]):
    if n_clicks:
        return not bool(is_color_dynamic)
    return dash.no_update


if __name__ == "__main__":
    import socket

    def get_ip(method: str = "broadcast") -> str:
        """Detect the IP address of this device."""
        s_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            if method == "broadcast":
                s_socket.connect(("10.255.255.255", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "udp":
                s_socket.connect(("8.8.8.8", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "host":
                ip_value = socket.gethostbyname(socket.gethostname())
            else:
                raise ConnectionError
        except Exception:  # pylint: disable=broad-except
            ip_value = "localhost"
        finally:
            s_socket.close()
        return ip_value

    app.run(host=get_ip(), port="8080", debug=False)



"""Pytest configurations
"""

from typing import TYPE_CHECKING
import pytest
import logging

import chromedriver_binary  # noqa: F401


def pytest_addoption(parser):
    parser.addoption(
        "--with-dash",
        action="store_true",
        help=(
            "Run the optional tests for local dash app. They require Chrome driver "
            "supports."
        ),
    )


def pytest_configure(config):
    logging.getLogger("werkzeug").setLevel(logging.ERROR)
    config.addinivalue_line(
        "markers", "with_dash: run the dash test with Chrome driver."
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--with-dash"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_wdash = pytest.mark.skip(reason="need --with-dash option to run")
    for item in items:
        if "with_dash" in item.keywords:
            item.add_marker(skip_wdash)


if not TYPE_CHECKING:
    try:
        from selenium.webdriver.chrome.options import Options

        def pytest_setup_options():
            # raise TypeError(same_hook)
            # The following configuration requires the installation of chrome!
            options = Options()
            # Removes a bunch of errors on Windows, like
            # USB: usb_device_win.cc:93 Failed to read descriptors from ...
            options.add_experimental_option("excludeSwitches", ["enable-logging"])
            # Do not need to configure --headless because it is included in the
            # pytest option (provided by dash_duo).
            # options.add_argument("--headless")
            options.add_argument("--no-sandbox")
            options.add_argument("--disable-gpu")
            options.add_argument("--window-size=1920,1080")
            return options

    except ImportError:
        pass



# -*- coding: UTF-8 -*-
"""
Minimal
=======
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

License
-------
MIT License

Description
-----------
The minimal demo for the basic functionalities.
"""

import os
import json

from typing import Optional

import dash
from dash import dcc, html
from dash import Output, Input


if __name__ == "__main__":
    import sys

    sys.path.append(os.path.dirname(os.path.dirname(__file__)))


import dash_picture_annotation as dpa


app = dash.Dash(
    __name__,
    assets_folder=os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "tests", "assets"
    ),
)

default_data = {
    "timestamp": 0,
    "data": [
        {
            "id": "TestID",
            "mark": {
                "x": 200.87731048580005,
                "y": 78.57834993258817,
                "width": 196.74579219762532,
                "height": 198.54529639455487,
                "type": "RECT",
            },
            "comment": "Test-init",
        },
    ],
}

app.layout = html.Div(
    [
        dcc.Loading(
            dpa.DashPictureAnnotation(
                id="annotator",
                style={"height": "80vh"},
                data=default_data,
                image="/assets/test_image.svg",
                options=None,
                clearable_dropdown=True,
            ),
            delay_show=1000,
        ),
        dcc.Loading(
            html.Div(
                id="output",
                children=json.dumps(default_data),
                style={"marginBottom": "1rem"},
            ),
            delay_show=1000,
        ),
    ]
)


@app.callback(
    Output("output", "children"), Input("annotator", "data"), prevent_initial_call=False
)
def get_annotation(data) -> Optional[str]:
    if data is None:
        return None
    return json.dumps(data)


if __name__ == "__main__":
    import socket

    def get_ip(method: str = "broadcast") -> str:
        """Detect the IP address of this device."""
        s_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            if method == "broadcast":
                s_socket.connect(("10.255.255.255", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "udp":
                s_socket.connect(("8.8.8.8", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "host":
                ip_value = socket.gethostbyname(socket.gethostname())
            else:
                raise ConnectionError
        except Exception:  # pylint: disable=broad-except
            ip_value = "localhost"
        finally:
            s_socket.close()
        return ip_value

    app.run(host=get_ip(), port="8080", debug=False)



# -*- coding: UTF-8 -*-
"""
Typehints
=========
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Extra typehints used by this project.
"""

import collections.abc

from typing import Union, Any, TypeVar

try:
    from typing import Sequence, Callable, Mapping
    from typing import List, Type
except ImportError:
    from collections.abc import Sequence, Callable, Mapping
    from builtins import list as List, type as Type

from typing_extensions import Literal, TypedDict, TypeGuard


T = TypeVar("T")

__all__ = (
    "AnnoMark",
    "AnnoItem",
    "Annotations",
    "AnnoStyle",
    "DashSelectOptionItem",
    "Size",
    "Scale",
    "NSAnnoItem",
    "NSAnnotations",
    "is_sequence_of",
    "is_anno_mark",
    "is_anno_item",
    "is_annotations",
    "is_dash_select_option_item",
)


class AnnoMark(TypedDict):
    """The `mark` property in the annotation item.

    This property contains the position and the shape of the bounding box. Currently,
    we only support the rectangular annotation.
    """

    x: float
    """The x (horizontal) position of the upper left corner of the annotation item."""

    y: float
    """The y (vertical) position of the upper left corner of the annotation item."""

    width: float
    """The width of the bounding box."""

    height: float
    """The height of the bounding box."""

    type: Literal["RECT"]
    """The type of the annotation shape. Currently, we only support `"RECT"`
    (rectangle)."""


class _AnnoItem(TypedDict):
    """Annotation item. (private, internal)

    The internal and incomplete definition of `AnnoItem`.
    """

    id: str
    """The ID of the annotation item. This ID is only used for locating which item
    is currently selected by users."""

    mark: AnnoMark
    """The boudning box and the shape information of this annotation item."""


class AnnoItem(_AnnoItem, total=False):
    """Annotation item.

    This dictionary represents the definition of one annotation item with its bounding
    box information.
    """

    comment: str
    """The text attached to this annotation item. Typically, this value is the type of
    the label."""


class _Annotations(TypedDict):
    """The collection of annotations. (private, internal)

    The internal and incomplete definition of `Annotations`.
    """

    data: List[AnnoItem]
    """A collection of annotation items. These items are decoded from JSON data and
    can be modified. Use these items to specify the annotations."""


class Annotations(_Annotations, total=False):
    """The collection of annotations.

    This dictionary contain the data of annotations.
    """

    timestamp: int
    """The time stamp recording when the user interaction changes the data. Note that
    this value is necessary for ensuring that the user triggered changes will not be
    omitted. From the server side, this value can be set by `0` because the user
    interaction will use a higher value to replace it."""


class AnnoStyle(TypedDict, total=False):
    """The css-styles of the annotation marker (box).

    If this value is specified as a string, the string will be parsed as the default
    color of the annotation boxes.
    """

    padding: float
    """Text padding."""

    fontSize: float
    """Text font size."""

    fontColor: str
    """Text font color."""

    fontBackground: str
    """Text background color."""

    fontFamily: str
    """Text font name."""

    lineWidth: float
    """Stroke width."""

    shapeBackground: str
    """Background color in the middle of the marker."""

    shapeStrokeStyle: str
    """Shape stroke color."""

    shadowBlur: float
    """Stroke shadow blur."""

    shapeShadowStyle: str
    """Stroke shape shadow color."""

    transformerBackground: str
    """Color of the scalable dots around the selected box."""

    transformerSize: float
    """Size of the scalable dots around the selected box."""


class _DashSelectOptionItem(TypedDict):
    """The option item of a Dash selectable component. (private, internal)

    The internal and incomplete definition of `DashSelectOptionItem`.
    """

    label: str
    """Label (displayed text) of the option."""

    value: Any
    """The value of the option which will be applied to the annotation data."""


class DashSelectOptionItem(_DashSelectOptionItem, total=False):
    """The option item of a Dash selectable component.

    The available options of the annotator. The usage is like the selector component
    `dcc.Dropdown(options=...)`. Each item represents an available choice of the
    annotation type.
    """

    disabled: bool
    """A flag. If specified, this option item will be not selectable."""


class Size(TypedDict, total=False):
    """The requirement of the minimal annotation size. Any newly created annotation
    with a size smaller than this size will be dropped.

    If this value is configured as a scalar, will use it for both `width` and `height`.

    If any of the value is not set or configured as invalid values, will use `0`.
    """

    width: float
    """Requirement of the minimal width of an annotation."""

    height: float
    """Requirement of the minimal height of an annotation."""


class Scale(TypedDict, total=False):
    """The initial image scale. This value can only be configured by users. The scaling
    reflected by the wheel event will not influence this value. Note that this value
    needs to be updated by a different value to make it take effect.
    """

    scale: float
    """The scale related to the initial scale of the annotated image. If not specified,
    will use `1.0`."""

    offset_x: float
    """The relative X offset. If not specified, will use `0.5` (center of the
    width)."""

    offset_y: float
    """The relative Y offset. If not specified, will use `0.5` (center of the
    height)."""

    timestamp: int
    """An optional timestamp value. This value will not be actually used, if it is
    configured, it can be used for letting the component know the scale should
    be updated."""


NSAnnoItem = Union[AnnoItem, AnnoMark, Mapping[str, Any]]
"""The type of a not sanitized annotation item."""


class _NSAnnotations(TypedDict):
    """The type of a not sanitized annotation collection. (private, internal)

    The internal and incomplete definition of `NSAnnotations`.
    """

    data: Sequence[NSAnnoItem]
    """A sequence of not sanitized data items. These items will be sanitized can
    put into a sanitized data item list."""


class NSAnnotations(_NSAnnotations, total=False):
    """The type of a not sanitized annotation collection."""


def is_sequence_of(
    data: Any, validator: Union[Type[T], Callable[[Any], TypeGuard[T]]]
) -> TypeGuard[Sequence[T]]:
    """Check whether `data` is `Sequence[T]`, where `T` is specified by `validator`."""
    if not isinstance(data, collections.abc.Sequence):
        return False
    if not data:
        return True
    if isinstance(validator, type):
        if isinstance(data, validator):
            return False
        return all(isinstance(ditem, validator) for ditem in data)
    else:
        return all(validator(ditem) for ditem in data)


def is_anno_mark(data: Any) -> TypeGuard[AnnoMark]:
    """Implementation of `isinstance(data, AnnoMark)`."""
    if not isinstance(data, collections.abc.Mapping):
        return False
    for key in ("x", "y", "width", "height", "type"):
        if key not in data:
            return False
    if data["type"] != "RECT":
        return False
    return True


def is_anno_item(data: Any) -> TypeGuard[AnnoItem]:
    """Implementation of `isinstance(data, AnnoItem)`."""
    if not isinstance(data, collections.abc.Mapping):
        return False
    for key in ("id", "mark"):
        if key not in data:
            return False
    if not is_anno_mark(data["mark"]):
        return False
    if not isinstance(data.get("comment", ""), str):
        return False
    return True


def is_annotations(data: Any) -> TypeGuard[Annotations]:
    """Implementation of `isinstance(data, Annotations)`."""
    if not isinstance(data, collections.abc.Mapping):
        return False
    timestamp = data.get("timestamp")
    if timestamp is not None and (not isinstance(timestamp, int)):
        return False
    anno_data = data.get("data")
    if not isinstance(anno_data, collections.abc.MutableSequence):
        return False
    for anno in anno_data:
        if not is_anno_item(anno):
            return False
    return True


def is_dash_select_option_item(obj: Any) -> TypeGuard[DashSelectOptionItem]:
    """Implementation of `isinstance(data, DashSelectOptionItem)`."""
    if not isinstance(obj, collections.abc.Mapping):
        return False
    if "label" in obj and "value" in obj:
        return True
    return False



# -*- coding: UTF-8 -*-
"""
Version
=======
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Version resolver for building the package. This package is only included in the wheel.
It is used for dynamically extracting the package version from `package.json`.
"""

import os
import json


__all__ = ("__version__",)

_basepath = "."
_filepath = os.path.abspath(os.path.join(_basepath, "package.json"))
with open(_filepath) as f:
    __version__ = str(json.load(f)["version"])



# -*- coding: UTF-8 -*-
"""
Usage
=====
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
A demo for the project. Run the following command to view the performance:
``` shell
python usage.py
```
"""

import os

from typing import Optional, Any

try:
    from typing import Sequence
except ImportError:
    from collections.abc import Sequence

import dash
import json
from dash import Dash, dcc, callback, html, Input, Output, State
import dash_picture_annotation as dpa

app = Dash(
    __name__, assets_folder=os.path.join(os.path.dirname(__file__), "tests", "assets")
)


default_data = {
    "timestamp": None,
    "data": [
        {
            "id": "AYAder",
            "mark": {
                "x": 200.87731048580005,
                "y": 78.57834993258817,
                "width": 196.74579219762532,
                "height": 198.54529639455487,
                "type": "RECT",
            },
            "comment": "Title",
        },
    ],
}

default_options = [
    {"value": "Title", "label": "Title"},
    {"value": "Label", "label": "Label"},
]

image_list = ["test_image.svg", "test_philips_PM5544.svg"]

styles = {
    "mr1": {"marginRight": "1rem"},
    "mb1": {"marginBottom": "1rem"},
}

app.layout = html.Div(
    [
        dcc.Loading(
            dpa.DashPictureAnnotation(
                id="annotator",
                style={"height": "80vh", "marginBottom": "1rem"},
                data=default_data,
                image="/assets/test_image.svg",
                options=default_options,
                clearable_dropdown=True,
            ),
            delay_show=1000,
        ),
        html.Div(
            (
                html.Button(
                    children="Toggle disabled", id="btn-disabled", style=styles["mr1"]
                ),
                html.Button(
                    children="Toggle color", id="btn-color", style=styles["mr1"]
                ),
                html.Button(
                    children="Toggle options", id="btn-options", style=styles["mr1"]
                ),
                html.Button(
                    children="Reset Annotations", id="btn-reset", style=styles["mr1"]
                ),
                html.Button(children="Set scale", id="btn-scale", style=styles["mr1"]),
                html.Button(
                    children="Change Image", id="btn-changeimg", style=styles["mr1"]
                ),
            ),
            style=styles["mb1"],
        ),
        dcc.Loading(
            html.Div(
                id="output", children=json.dumps(default_data), style=styles["mb1"]
            ),
            delay_show=1000,
        ),
    ]
)


@callback(
    Output("annotator", "disabled"),
    Input("btn-disabled", "n_clicks"),
    State("annotator", "disabled"),
)
def toggle_disabled(n_clicks: Optional[int], disabled: Optional[bool]):
    if n_clicks:
        return not bool(disabled)
    return dash.no_update


@callback(
    Output("annotator", "is_color_dynamic"),
    Input("btn-color", "n_clicks"),
    State("annotator", "is_color_dynamic"),
)
def toggle_color(n_clicks: Optional[int], is_color_dynamic: Optional[bool]):
    if n_clicks:
        return not bool(is_color_dynamic)
    return dash.no_update


@callback(
    Output("annotator", "options"),
    Input("btn-options", "n_clicks"),
    State("annotator", "options"),
)
def toggle_options(n_clicks: Optional[int], options: Optional[Sequence[Any]]):
    if n_clicks:
        if options:
            return None
        else:
            return default_options
    return dash.no_update


@app.callback(
    Output("annotator", "data"),
    Input("btn-reset", "n_clicks"),
    prevent_initial_call=False,
)
def reset_data(n_clicks: Optional[int]):
    if n_clicks:
        return default_data
    return dash.no_update


@callback(
    Output("annotator", "init_scale"),
    Input("btn-scale", "n_clicks"),
)
def set_scale(n_clicks: Optional[int]):
    if n_clicks:
        return dpa.sanitize_scale(
            scale=1.0,
            offset_x=0.5,  # 0: left, 0.5: center, 1.0: right
        )
    return dash.no_update


@callback(
    Output("annotator", "image"),
    Input("btn-changeimg", "n_clicks"),
    State("annotator", "image"),
    prevent_initial_call=False,
)
def reset_img(n_clicks: Optional[int], prev_image: Optional[str]) -> str:
    if n_clicks:
        if isinstance(prev_image, str):
            try:
                idx = image_list.index(os.path.split(prev_image)[-1])
            except ValueError:
                return "/assets/test_image.svg"
            next_idx = (idx + 1) % len(image_list)
            return "/assets/{0}".format(image_list[next_idx])
    return "/assets/test_image.svg"


@callback(
    Output("output", "children"), Input("annotator", "data"), prevent_initial_call=False
)
def get_annotation(data) -> str:
    return json.dumps(data)


if __name__ == "__main__":
    import socket

    def get_ip(method: str = "broadcast") -> str:
        """Detect the IP address of this device."""
        s_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            if method == "broadcast":
                s_socket.connect(("10.255.255.255", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "udp":
                s_socket.connect(("8.8.8.8", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "host":
                ip_value = socket.gethostbyname(socket.gethostname())
            else:
                raise ConnectionError
        except Exception:  # pylint: disable=broad-except
            ip_value = "localhost"
        finally:
            s_socket.close()
        return ip_value

    app.run(host=get_ip(), port="8080", debug=False)



# -*- coding: UTF-8 -*-
"""
Size Control
============
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

License
-------
MIT License

Description
-----------
The demo for controlling the minimal size of annotations.
"""

import os
import json

from typing import Optional

import dash
from dash import dcc, html
from dash import Output, Input


if __name__ == "__main__":
    import sys

    sys.path.append(os.path.dirname(os.path.dirname(__file__)))


import dash_picture_annotation as dpa


app = dash.Dash(
    __name__,
    assets_folder=os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "tests", "assets"
    ),
)

app.layout = html.Div(
    [
        dcc.Loading(
            dpa.DashPictureAnnotation(
                id="annotator",
                style={"height": "80vh"},
                data=None,
                image="/assets/test_image.svg",
                options=None,
                clearable_dropdown=True,
                size_minimal=20,
            ),
            delay_show=1000,
        ),
        dcc.Loading(
            html.Div(id="output", children=None, style={"marginBottom": "1rem"}),
            delay_show=1000,
        ),
    ]
)


@app.callback(
    Output("output", "children"), Input("annotator", "data"), prevent_initial_call=False
)
def get_annotation(data) -> Optional[str]:
    if data is None:
        return None
    return json.dumps(data)


if __name__ == "__main__":
    import socket

    def get_ip(method: str = "broadcast") -> str:
        """Detect the IP address of this device."""
        s_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            if method == "broadcast":
                s_socket.connect(("10.255.255.255", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "udp":
                s_socket.connect(("8.8.8.8", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "host":
                ip_value = socket.gethostbyname(socket.gethostname())
            else:
                raise ConnectionError
        except Exception:  # pylint: disable=broad-except
            ip_value = "localhost"
        finally:
            s_socket.close()
        return ip_value

    app.run(host=get_ip(), port="8080", debug=False)



# -*- coding: UTF-8 -*-
"""
States
======
@ Dash Picture Annotation - Tests

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
The tests for the `states.py` application. These tests will run a browser emulator
powered by `selenium` and `dash.testing`. The basic functionalities of the demo
will be checked one by one.
"""

import json
import logging

try:
    from typing import Generator
except ImportError:
    from collections.abc import Generator

import pytest

import dash
from dash.testing.application_runners import import_app
from dash.testing.composite import DashComposite
from selenium.webdriver.remote.webelement import WebElement

from .utils import wait_for_dcc_loading, asset_folder
from .imgutils import decode_base64_image, compare_images, ImageLoader
from .actions import drag_and_drop, get_canvas_image


__all__ = ("TestStates",)


@pytest.mark.with_dash
class TestStates:
    """Test the rendering the Dash app: States"""

    @pytest.fixture(scope="class")
    def dash_app(self) -> Generator[dash.Dash, None, None]:
        """Fixture of the dash application."""
        log = logging.getLogger("dash_picture_annotation.test")
        log.info("Initialize the Dash app.")
        app = import_app("examples.states")
        yield app
        log.info("Remove the Dash app.")
        del app

    @pytest.fixture(scope="class")
    def ref_images(self) -> Generator[ImageLoader, None, None]:
        """Fixture of the testing images."""
        log = logging.getLogger("dash_picture_annotation.test")
        log.info("Use the reference images.")
        yield ImageLoader(root=asset_folder())

    def test_states_toggle_disabled(
        self, dash_duo: DashComposite, dash_app: dash.Dash
    ) -> None:
        """Test the states app and toggle the `disabled` property of the annotator..

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_disabled: WebElement = dash_duo.find_element("#btn-disabled")
        output: WebElement = dash_duo.find_element("#output")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))

        wait_for_dcc_loading(dash_duo, "#output")
        annotation_ids = tuple(
            anno["id"] for anno in json.loads(output.text).get("data", tuple())
        )
        log.info("Updated annotation IDs: {0}".format(annotation_ids))
        assert len(annotation_ids) == 2

        btn_disabled.click()
        dash_duo.wait_for_element_by_css_selector(
            "#annotator > div[class^='dpa-overlay-container']"
        )
        drag_and_drop(dash_duo, annotator, pos_start=(400, 100), pos_end=(600, 300))
        wait_for_dcc_loading(dash_duo, "#output")
        annotation_ids = tuple(
            anno["id"] for anno in json.loads(output.text).get("data", tuple())
        )
        log.info("Updated annotation IDs: {0}".format(annotation_ids))
        assert len(annotation_ids) == 2

        btn_disabled.click()
        dash_duo.wait_for_element_by_css_selector("#annotator > .rp-stage > .rp-image")
        drag_and_drop(dash_duo, annotator, pos_start=(400, 100), pos_end=(600, 300))
        wait_for_dcc_loading(dash_duo, "#output")
        annotation_ids = tuple(
            anno["id"] for anno in json.loads(output.text).get("data", tuple())
        )
        log.info("Updated annotation IDs: {0}".format(annotation_ids))
        assert len(annotation_ids) == 3

    def test_states_toggle_image(
        self,
        dash_duo: DashComposite,
        dash_app: dash.Dash,
        ref_images: ImageLoader,
    ) -> None:
        """Test the states app and toggle the `image` property of the annotator.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_image: WebElement = dash_duo.find_element("#btn-image")
        output: WebElement = dash_duo.find_element("#output")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))

        wait_for_dcc_loading(dash_duo, "#output")
        anno_ids_before = tuple(
            anno["id"] for anno in json.loads(output.text).get("data", tuple())
        )
        log.info("Current annotation IDs: {0}".format(anno_ids_before))
        assert len(anno_ids_before) == 2

        for img_name in ("test_philips_PM5544", "test_image"):
            btn_image.click()
            res = get_canvas_image(dash_duo, annotator)
            img = decode_base64_image(res)
            ssim = compare_images(img, ref_images["ref_" + img_name])
            log.info('SSIM of the image "{0}": {1}'.format(img_name, ssim))

            anno_ids = tuple(
                anno["id"] for anno in json.loads(output.text).get("data", tuple())
            )
            log.info("Current annotation IDs: {0}".format(anno_ids))
            assert len(anno_ids) == 2
            assert len(set(anno_ids).difference(anno_ids_before)) == 0
            assert ssim > 0.85

    def test_states_reset_data(
        self,
        dash_duo: DashComposite,
        dash_app: dash.Dash,
    ) -> None:
        """Test the states app and reset the data of the annotator from the server side.

        The release of the mouse will trigger the callbacks.
        """
        log = logging.getLogger("dash_picture_annotation.test")

        # Start a dash app contained as the variable `app` in `usage.py`
        dash_duo.start_server(dash_app)

        annotator: WebElement = dash_duo.find_element("#annotator")
        btn_data: WebElement = dash_duo.find_element("#btn-data")
        output: WebElement = dash_duo.find_element("#output")

        drag_and_drop(dash_duo, annotator, pos_start=(100, 100), pos_end=(300, 300))
        drag_and_drop(dash_duo, annotator, pos_start=(400, 100), pos_end=(600, 300))

        wait_for_dcc_loading(dash_duo, "#output")
        anno_ids_before = tuple(
            anno["id"] for anno in json.loads(output.text).get("data", tuple())
        )
        log.info("Current annotation IDs: {0}".format(anno_ids_before))
        assert len(anno_ids_before) == 3

        btn_data.click()
        wait_for_dcc_loading(dash_duo, "#output")
        annotations = json.loads(output.text).get("data", tuple())
        anno_ids = tuple(anno["id"] for anno in annotations)
        log.info("Current annotation IDs: {0}".format(anno_ids))
        assert len(anno_ids) == 1
        assert annotations[0]["id"] == "TestID"



# -*- coding: UTF-8 -*-
"""
Setup
=====

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

Description
-----------
Dummy `setup.py` for ensuring the compatibility. The configurations are made in
`pyproject.toml`.

Run the following commands to build the package/wheel.
``` shell
python -m build
```
"""

from setuptools import setup

setup()



# -*- coding: UTF-8 -*-
"""
States
======
@ Dash Picture Annotation

Author
------
Yuchen Jin (cainmagi)
cainmagi@gmail.com

License
-------
MIT License

Description
-----------
The demo for updating the annotator states from the server side.
"""

import os
import json

from typing import Optional

import dash
from dash import dcc, html
from dash import Output, Input, State


if __name__ == "__main__":
    import sys

    sys.path.append(os.path.dirname(os.path.dirname(__file__)))


import dash_picture_annotation as dpa


app = dash.Dash(
    __name__,
    assets_folder=os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "tests", "assets"
    ),
)

image_list = ["test_image.svg", "test_philips_PM5544.svg"]

default_data = {
    "timestamp": 0,
    "data": [
        {
            "id": "TestID",
            "mark": {
                "x": 200.87731048580005,
                "y": 78.57834993258817,
                "width": 196.74579219762532,
                "height": 198.54529639455487,
                "type": "RECT",
            },
            "comment": "Test-init",
        },
    ],
}

styles = {
    "mb1": {"marginBottom": "1rem"},
    "mr1": {"marginRight": "1rem"},
}

app.layout = html.Div(
    [
        dcc.Loading(
            dpa.DashPictureAnnotation(
                id="annotator",
                style={"height": "80vh"},
                data=default_data,
                image=None,
                options=None,
                clearable_dropdown=True,
            ),
            delay_show=1000,
        ),
        dcc.Loading(
            html.Div(
                id="output", children=json.dumps(default_data), style=styles["mb1"]
            ),
            delay_show=1000,
        ),
        html.Div(
            children=(
                html.Button(
                    children="Toggle disabled", id="btn-disabled", style=styles["mr1"]
                ),
                html.Button(
                    children="Toggle image", id="btn-image", style=styles["mr1"]
                ),
                html.Button(children="Reset data", id="btn-data"),
            ),
            style=styles["mb1"],
        ),
    ]
)


@app.callback(
    Output("annotator", "disabled"),
    Input("btn-disabled", "n_clicks"),
    State("annotator", "disabled"),
    prevent_initial_call=True,
)
def toggle_disabled(n_clicks: Optional[int], is_disabled: Optional[bool]):
    if n_clicks:
        return not bool(is_disabled)
    return dash.no_update


@app.callback(
    Output("annotator", "image"),
    Input("btn-image", "n_clicks"),
    State("annotator", "image"),
    prevent_initial_call=False,
)
def toggle_image(n_clicks: Optional[int], prev_image: Optional[str]) -> str:
    if n_clicks:
        if isinstance(prev_image, str):
            try:
                idx = image_list.index(os.path.split(prev_image)[-1])
            except ValueError:
                return "/assets/test_image.svg"
            next_idx = (idx + 1) % len(image_list)
            return "/assets/{0}".format(image_list[next_idx])
    return "/assets/test_image.svg"


@app.callback(
    Output("annotator", "data"),
    Input("btn-data", "n_clicks"),
    prevent_initial_call=False,
)
def reset_data(n_clicks: Optional[int]):
    if n_clicks:
        return default_data
    return dash.no_update


@app.callback(
    Output("output", "children"), Input("annotator", "data"), prevent_initial_call=False
)
def get_annotation(data) -> Optional[str]:
    if data is None:
        return None
    return json.dumps(data)


if __name__ == "__main__":
    import socket

    def get_ip(method: str = "broadcast") -> str:
        """Detect the IP address of this device."""
        s_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            if method == "broadcast":
                s_socket.connect(("10.255.255.255", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "udp":
                s_socket.connect(("8.8.8.8", 1))
                ip_value = s_socket.getsockname()[0]
            elif method == "host":
                ip_value = socket.gethostbyname(socket.gethostname())
            else:
                raise ConnectionError
        except Exception:  # pylint: disable=broad-except
            ip_value = "localhost"
        finally:
            s_socket.close()
        return ip_value

    app.run(host=get_ip(), port="8080", debug=False)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class DashPictureAnnotation(Component):
    """A DashPictureAnnotation component.
DashPictureAnnotation is a Dash porting version for the React component:
`react-picture-annotation/ReactPictureAnnotation`.

This component provides a annotator that allows users to create, modify, or delete
the annotation information for a specific picture. This dash version has been
upgraded by adding the following features:
1. Responsive size with respect to the parent component.
2. Annotation type specified by a selector rather than an input box.
3. Only trigger the data update when the mouse is released.
4. Extensive functionalities for fine-grained customization of colors.
5. Disabling the annotator by a flag.
6. Setting a lower boundary of the annotation size to prevent small annotations
   created by mistake.

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- class_name (string; optional):
    The css-class of the component. Use ` ` to separate different
    names. Often used with CSS to style elements with common
    properties.

- clearable_dropdown (boolean; default False):
    A flag. Set it to allow the annotation comment to be cleared when
    the dropdown box is being used.

- colors (dict with strings as keys and values of type string; optional):
    A dictionary of colors. The keys and the values represent the
    annotation name (specified by the `comment` property of `data`)
    and the color string, respectively. This method is used for
    specifying the annotation colors manually. It has a higher
    priority compared to `is_color_dynamic`.

- data (dict; optional):
    The annotation data. This value is a sequence of annotation items.
    Each item contain the positional information and the annotation
    comment (the object type). This value will can be updated by
    callbacks or user operations on the annotator.

    `data` is a dict with keys:

    - timestamp (number; optional):
        The time stamp of the current data. This value is used for
        identify the version of the data. It can also notify dash that
        the callback should be fired when this value changes.

    - data (list of dicts; optional):
        The body of the data. Returned from the React annotation data.

        `data` is a list of dicts with keys:

        - id (string; optional):

            The ID of the annotation item. It is only used for locating

            which annotation item is selected by users.

        - mark (dict; optional):

            The bounding box information of the annotation item.

            `mark` is a dict with keys:

            - x (number; optional):

                The X (horizontal) position of the upper left corner.

            - y (number; optional):

                The Y (vertical) position of the upper left corner.

            - width (number; optional):

                The width of the annotation item.

            - height (number; optional):

                The height of the annotation item.

            - type (a value equal to: "RECT"; optional):

                The shape of the annotation item. Currently, we only

                support \"RECT\".

        - comment (string; optional):

            The text comment on this annotation item. Typically, this

            value is specified by the type of the label.

- disabled (boolean; default False):
    A flag for disabling the annotator (make unclickable).

- image (string; optional):
    The URL to the image to be currently displayed on the annotator.
    The usage is similar to the property of the HTML image tag: `<img
    src={...}>`.

- init_scale (dict; default 1.0):
    The initial image scale. This value can only be configured by
    users. The scaling reflected by the wheel event will not influence
    this value. Note that this value needs to be updated by a
    different value to make it take effect.

    `init_scale` is a number | dict with keys:

    - scale (number; optional):
        The scale related to the initial scale of the annotated image.
        If not specified, will use `1.0`.

    - offset_x (number; optional):
        The relative X offset. If not specified, will use `0.5`
        (center of the width).

    - timestamp (number; optional):
        The relative Y offset. If not specified, will use `0.5`
        (center of the height).

- is_color_dynamic (boolean; default False):
    A flag. If this flag is turned on, will make the color of each
    annotation box dynamically calculated based on the text of the
    annotation. An annotation box without a text comment will not be
    influenced.

- loading_state (dict; optional):
    Object that holds the loading state object coming from
    dash-renderer.

    `loading_state` is a dict with keys:

    - is_loading (boolean; optional):
        Determines if the component is loading or not.

    - prop_name (string; optional):
        Holds which property is loading.

    - component_name (string; optional):
        Holds the name of the component that is loading.

- options (list of dicts; optional):
    The available options of the annotator. The usage is like the
    selector component `dcc.Dropdown(options=...)`. Each item
    represents an available choice of the annotation type. If this
    value is empty, will use a input box that allows users to specify
    any type names.

    `options` is a list of dicts with keys:

    - label (string; required):
        Label (displayed text) of the option.

    - value (boolean | number | string | dict | list; required):
        The value of the option which will be applied to the
        annotation data.

    - disabled (boolean; optional):
        A flag. If specified, this option item will be not selectable. | list of string | number | booleans | dict | a value equal to: null

- placeholder_dropdown (string; default "Select a tag"):
    The placeholder text when the editor is the dropdown component
    (the property `options` contains valid items).

- placeholder_input (string; default "Input tag here"):
    The placeholder text when the editor is the input box (the
    property `options` is empty).

- selected_id (string | a value equal to: null; optional):
    The ID of the currently selected annotation. This property is
    read-only. It will be automatically set when users select an
    annotation. A valid ID is a string. When no annotation is
    selected, this property returns `None`.

- size_minimal (dict; default {  width: 0,  height: 0,}):
    The requirement of the minimal annotation size. Any newly created
    annotation with a size smaller than this size will be dropped. If
    this value is configured as a scalar, will use it for both `width`
    and `height`. If any of the value is not set or configured as
    invalid values, will use `0`.

    `size_minimal` is a number | dict with keys:

    - width (number; optional):
        Requirement of the minimal width of an annotation.

    - height (number; optional):
        Requirement of the minimal height of an annotation.

- style (dict; default {height: "60vh"}):
    The css-styles which will override styles of the component
    container.

- style_annotation (dict; optional):
    The css-styles of the annotation marker (box). If this value is
    specified as a string, the string will be parsed as the default
    color of the annotation boxes.

    `style_annotation` is a string | dict with keys:

    - padding (number; optional):
        Shape style: text padding.

    - fontSize (number; optional):
        Shape style: text font size.

    - fontColor (string; optional):
        Shape style: text font color.

    - fontBackground (string; optional):
        Shape style: text background color.

    - fontFamily (string; optional):
        Shape style: text font name.

    - lineWidth (number; optional):
        Shape style: stroke width.

    - shapeBackground (string; optional):
        Shape style: background color in the middle of the marker.

    - shapeStrokeStyle (string; optional):
        Shape style: shape stroke color.

    - shadowBlur (number; optional):
        Shape style: stroke shadow blur.

    - shapeShadowStyle (string; optional):
        Shape style: shape shadow color.

    - transformerBackground (string; optional):
        Shape style: color of the scalable dots around the selected
        box.

    - transformerSize (number; optional):
        Shape style: size of the scalable dots around the selected
        box."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_picture_annotation'
    _type = 'DashPictureAnnotation'
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, class_name=Component.UNDEFINED, style=Component.UNDEFINED, style_annotation=Component.UNDEFINED, colors=Component.UNDEFINED, image=Component.UNDEFINED, data=Component.UNDEFINED, options=Component.UNDEFINED, selected_id=Component.UNDEFINED, placeholder_input=Component.UNDEFINED, placeholder_dropdown=Component.UNDEFINED, clearable_dropdown=Component.UNDEFINED, disabled=Component.UNDEFINED, is_color_dynamic=Component.UNDEFINED, init_scale=Component.UNDEFINED, size_minimal=Component.UNDEFINED, loading_state=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'class_name', 'clearable_dropdown', 'colors', 'data', 'disabled', 'image', 'init_scale', 'is_color_dynamic', 'loading_state', 'options', 'placeholder_dropdown', 'placeholder_input', 'selected_id', 'size_minimal', 'style', 'style_annotation']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'class_name', 'clearable_dropdown', 'colors', 'data', 'disabled', 'image', 'init_scale', 'is_color_dynamic', 'loading_state', 'options', 'placeholder_dropdown', 'placeholder_input', 'selected_id', 'size_minimal', 'style', 'style_annotation']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(DashPictureAnnotation, self).__init__(**args)



"""
DO NOT MODIFY
This file is used to validate your publish settings.
"""
from __future__ import print_function

import os
import sys
import importlib


components_package = 'dash_picture_annotation'

components_lib = importlib.import_module(components_package)

missing_dist_msg = 'Warning {} was not found in `{}.__init__.{}`!!!'
missing_manifest_msg = '''
Warning {} was not found in `MANIFEST.in`!
It will not be included in the build!
'''

with open('MANIFEST.in', 'r') as f:
    manifest = f.read()


def check_dist(dist, filename):
    # Support the dev bundle.
    if filename.endswith('dev.js'):
        return True

    return any(
        filename in x
        for d in dist
        for x in (
            [d.get('relative_package_path')]
            if not isinstance(d.get('relative_package_path'), list)
            else d.get('relative_package_path')
        )
    )


def check_manifest(filename):
    return filename in manifest


def check_file(dist, filename):
    if not check_dist(dist, filename):
        print(
            missing_dist_msg.format(filename, components_package, '_js_dist'),
            file=sys.stderr
        )
    if not check_manifest(filename):
        print(missing_manifest_msg.format(filename),
              file=sys.stderr)


for cur, _, files in os.walk(components_package):
    for f in files:

        if f.endswith('js'):
            # noinspection PyProtectedMember
            check_file(components_lib._js_dist, f)
        elif f.endswith('css'):
            # noinspection PyProtectedMember
            check_file(components_lib._css_dist, f)
        elif not f.endswith('py'):
            check_manifest(f)
